# Authors: Isak Samsten
# License: BSD 3 clause

import math
import numbers
import warnings
from abc import ABCMeta, abstractmethod
from numbers import Integral, Real

import numpy as np
from joblib import Parallel, delayed
from scipy import sparse
from sklearn.base import OutlierMixin
from sklearn.ensemble import BaggingClassifier as SklearnBaggingClassifier
from sklearn.ensemble import BaggingRegressor as SklearnBaggingRegressor
from sklearn.ensemble._bagging import BaseBagging as SklearnBaseBagging
from sklearn.preprocessing import OneHotEncoder
from sklearn.utils import check_random_state, compute_sample_weight
from sklearn.utils._param_validation import HasMethods, Interval, StrOptions
from sklearn.utils.validation import check_is_fitted, check_scalar

from ..base import BaseEstimator
from ..transform._shapelet import ShapeletMixin
from ..tree import (
    ExtraShapeletTreeClassifier,
    ExtraShapeletTreeRegressor,
    IntervalTreeClassifier,
    IntervalTreeRegressor,
    PivotTreeClassifier,
    ProximityTreeClassifier,
    ShapeletTreeClassifier,
    ShapeletTreeRegressor,
)
from ..tree._tree import RocketTreeClassifier, RocketTreeRegressor


class BaseBagging(BaseEstimator, SklearnBaseBagging, metaclass=ABCMeta):
    """Base estimator for Wildboar ensemble estimators."""

    _parameter_constraints: dict = {
        "estimator": [HasMethods(["fit", "predict"]), None],
        "n_estimators": [Interval(Integral, 1, None, closed="left")],
        "max_samples": [
            Interval(Integral, 1, None, closed="left"),
            Interval(Real, 0, 1, closed="right"),
        ],
        "bootstrap": ["boolean"],
        "oob_score": ["boolean"],
        "warm_start": ["boolean"],
        "n_jobs": [None, Integral],
        "random_state": ["random_state"],
        "verbose": ["verbose"],
        "base_estimator": [
            HasMethods(["fit", "predict"]),
            StrOptions({"deprecated"}),
            None,
        ],
    }

    @abstractmethod
    def __init__(
        self,
        estimator=None,
        n_estimators=10,
        *,
        max_samples=1.0,
        bootstrap=True,
        oob_score=False,
        warm_start=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        base_estimator="deprecated",
    ):
        if base_estimator != "deprecated":
            warnings.warn(
                "base_estimator is deprecated and will be removed in 1.4",
                DeprecationWarning,
            )
            if estimator is None:
                estimator = base_estimator

        self.base_estimator = base_estimator
        super().__init__(
            estimator=estimator,
            n_estimators=n_estimators,
            max_samples=max_samples,
            max_features=1.0,
            bootstrap=bootstrap,
            bootstrap_features=False,
            oob_score=oob_score,
            warm_start=warm_start,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
        )

    def _validate_data(
        self,
        x="no_validation",
        y="no_validation",
        reset=True,
        validate_separately=False,
        **check_params,
    ):
        new_check_params = {"allow_3d": True}
        if "y_numeric" in check_params:
            new_check_params["y_numeric"] = check_params["y_numeric"]

        out = super()._validate_data(
            x,
            y,
            reset=reset,
            validate_separately=validate_separately,
            dtype=float,
            **new_check_params,
        )
        no_val_X = isinstance(x, str) and x == "no_validation"
        no_val_y = y is None or isinstance(y, str) and y == "no_validation"
        if no_val_X:
            return out
        elif no_val_y:
            x = out
        else:
            x, y = out

        if x.ndim == 3:
            x = x.reshape(x.shape[0], -1)

        if no_val_y:
            return x
        else:
            return x, y

    def fit(self, x, y, sample_weight=None):
        self._validate_params()
        x, y = self._validate_data(x, y, allow_3d=True, dtype=float)
        self._fit(
            x, y, self.max_samples, self.max_depth, sample_weight, check_input=False
        )
        return self

    def _make_estimator(self, append=True, random_state=None):
        estimator = super()._make_estimator(append, random_state)
        if self.n_dims_in_ > 1:
            estimator._force_n_dims = self.n_dims_in_
        return estimator

    def _more_tags(self):
        return {"X_types": ["2darray", "3darray"]}


class ForestMixin:
    """
    Mixin for tree based ensembles.
    """

    _parameter_constraints: dict = {
        "n_estimators": [Interval(Integral, 1, None, closed="left")],
        "bootstrap": ["boolean"],
        "oob_score": ["boolean"],
        "n_jobs": [Integral, None],
        "random_state": ["random_state"],
        "warm_start": ["boolean"],
    }

    def apply(self, x):
        check_is_fitted(self)
        x = self._validate_data(x, allow_3d=True, reset=False)
        results = Parallel(
            n_jobs=self.n_jobs,
            verbose=self.verbose,
            **{"prefer": "threads"},
        )(delayed(tree.apply)(x, check_input=False) for tree in self.estimators_)

        return np.array(results).T

    def decision_path(self, x):
        check_is_fitted(self)
        x = self._validate_data(x, allow_3d=True, reset=False)
        indicators = Parallel(
            n_jobs=self.n_jobs,
            verbose=self.verbose,
            **{"prefer": "threads"},
        )(
            delayed(tree.decision_path)(x, check_input=False)
            for tree in self.estimators_
        )

        n_nodes = [0]
        n_nodes.extend([i.shape[1] for i in indicators])
        n_nodes_ptr = np.array(n_nodes).cumsum()

        return sparse.hstack(indicators).tocsr(), n_nodes_ptr


class BaggingClassifier(BaseBagging, SklearnBaggingClassifier):
    """
    A bagging classifier.

    A bagging regressor is a meta-estimator that fits base classifiers
    on random subsets of the original data.

    Parameters
    ----------
    estimator : object, optional
        Base estimator of the ensemble. If `None`, the base estimator
        is a :class:`~wildboar.tree.ShapeletTreeRegressor`.
    n_estimators : int, optional
        The number of base estimators in the ensemble.
    max_samples : int or float, optional
        The number of samples to draw from `X` to train each base estimator.

        - if `int`, then draw `max_samples` samples.
        - if `float`, then draw `max_samples * n_samples` samples.
    bootstrap : bool, optional
        If the samples are drawn with replacement.
    oob_score : bool, optional
        Use out-of-bag samples to estimate generalization performance. Requires
        `bootstrap=True`.
    class_weight : dict or "balanced", optional
        Weights associated with the labels

        - if `dict`, weights on the form `{label: weight}`.
        - if "balanced" each class weight inversely proportional to
          the class frequency.
        - if `None`, each class has equal weight.
    warm_start : bool, optional
        When set to `True`, reuse the solution of the previous call
        to fit and add more estimators to the ensemble, otherwise, just fit
        a whole new ensemble.
    n_jobs : int, optional
        The number of jobs to run in parallel. A value of `None` means
        using a single core and a value of `-1` means using all cores.
        Positive integers mean the exact number of cores.
    random_state : int or RandomState, optional
        Controls the random resampling of the original dataset.

        - If `int`, `random_state` is the seed used by the
          random number generator.
        - If :class:`numpy.random.RandomState` instance, `random_state` is
          the random number generator.
        - If `None`, the random number generator is the
          :class:`numpy.random.RandomState` instance used by
          :func:`numpy.random`.
    verbose : int, optional
        Controls the output to standard error while fitting and predicting.
    base_estimator : object, optional
        Use `estimator` instead.

        .. deprecated:: 1.2
            `base_estimator` has been deprecated and will be removed in 1.4.
            Use `estimator` instead.
    """

    _parameter_constraints: dict = {**BaseBagging._parameter_constraints}

    def __init__(
        self,
        estimator=None,
        n_estimators=10,
        *,
        max_samples=1.0,
        bootstrap=True,
        oob_score=False,
        class_weight=None,
        warm_start=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        base_estimator="deprecated",
    ):
        super().__init__(
            estimator=ShapeletTreeClassifier() if estimator is None else estimator,
            n_estimators=n_estimators,
            max_samples=max_samples,
            bootstrap=bootstrap,
            oob_score=oob_score,
            warm_start=warm_start,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            base_estimator=base_estimator,
        )
        self.class_weight = class_weight

    def fit(self, x, y, sample_weight=None):
        self._validate_params()
        x, y = self._validate_data(x, y, allow_3d=True, dtype=float)

        if self.class_weight is not None:
            class_weight = compute_sample_weight(self.class_weight, y)
            if sample_weight is not None:
                sample_weight = sample_weight * class_weight
            else:
                sample_weight = class_weight

        self._fit(
            x, y, self.max_samples, self.max_depth, sample_weight, check_input=True
        )
        return self


class BaseForestClassifier(ForestMixin, BaggingClassifier, metaclass=ABCMeta):
    @abstractmethod
    def __init__(
        self,
        estimator,
        n_estimators=100,
        estimator_params=tuple(),
        *,
        oob_score=False,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        criterion="entropy",
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        class_weight=None,
        random_state=None,
    ):
        super().__init__(
            estimator=estimator,
            n_estimators=n_estimators,
            bootstrap=bootstrap,
            n_jobs=n_jobs,
            warm_start=warm_start,
            random_state=random_state,
            oob_score=oob_score,
        )
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.criterion = criterion
        self.min_samples_leaf = min_samples_leaf
        self.min_impurity_decrease = min_impurity_decrease
        self.class_weight = class_weight
        self.estimator_params = estimator_params

    def _fit(
        self,
        X,
        y,
        max_samples=None,
        max_depth=None,
        sample_weight=None,
        check_input=False,
    ):
        return super()._fit(X, y, max_samples, max_depth, sample_weight, False)

    def _parallel_args(self):
        return {"prefer": "threads"}


class BaseShapeletForestClassifier(BaseForestClassifier, metaclass=ABCMeta):
    """Base class for shapelet forest classifiers.

    Warnings
    --------
    This class should not be used directly. Use derived classes instead.
    """

    @abstractmethod
    def __init__(  # noqa: PLR0913
        self,
        estimator,
        n_estimators=100,
        estimator_params=tuple(),
        oob_score=False,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        n_shapelets="log2",
        min_shapelet_size=0.0,
        max_shapelet_size=1.0,
        metric="euclidean",
        metric_params=None,
        criterion="entropy",
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        class_weight=None,
        random_state=None,
    ):
        super().__init__(
            estimator=estimator,
            estimator_params=estimator_params,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
            criterion=criterion,
            class_weight=class_weight,
            n_estimators=n_estimators,
            bootstrap=bootstrap,
            n_jobs=n_jobs,
            warm_start=warm_start,
            random_state=random_state,
            oob_score=oob_score,
        )
        self.n_shapelets = n_shapelets
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.metric = metric
        self.metric_params = metric_params


class ShapeletForestClassifier(BaseShapeletForestClassifier):
    """
    An ensemble of random shapelet tree classifiers.

    A forest of randomized shapelet trees.

    Parameters
    ----------
    n_estimators : int, optional
        The number of estimators.
    n_shapelets : int, optional
        The number of shapelets to sample at each node.
    max_depth : int, optional
        The maximum depth of the tree. If `None` the tree is
        expanded until all leaves are pure or until all leaves contain less
        than `min_samples_split` samples.
    min_samples_split : int, optional
        The minimum number of samples to split an internal node.
    min_samples_leaf : int, optional
        The minimum number of samples in a leaf.
    min_impurity_decrease : float, optional
        A split will be introduced only if the impurity decrease is larger
        than or equal to this value.
    min_shapelet_size : float, optional
        The minimum length of a shapelets expressed as a fraction of
        *n_timestep*.
    max_shapelet_size : float, optional
        The maximum length of a shapelets expressed as a fraction of
        *n_timestep*.
    alpha : float, optional
        Dynamically decrease the number of sampled shapelets at each node
        according to the current depth, i.e.:

        ::
            w = 1 - exp(-abs(alpha) * depth)

        - if `alpha < 0`, the number of sampled shapelets decrease from
          `n_shapelets` towards 1 with increased depth.
        - if `alpha > 0`, the number of sampled shapelets increase from `1`
          towards `n_shapelets` with increased depth.
        - if `None`, the number of sampled shapelets are the same
          independeth of depth.
    metric : str or list, optional
        The distance metric.

        - If `str`, the distance metric used to identify the best
          shapelet.
        - If `list`, multiple metrics specified as a list of tuples,
          where the first element of the tuple is a metric name and the second
          element a dictionary with a parameter grid specification. A parameter
          grid specification is a dict with two mandatory and one optional
          key-value pairs defining the lower and upper bound on the values and
          number of values in the grid. For example, to specifiy a grid over
          the argument `r` with 10 values in the range 0 to 1, we would give
          the following specification: `dict(min_r=0, max_r=1, num_r=10)`.

          Read more about metric specifications in the `User guide
          <metric_specification>`__.

        .. versionchanged:: 1.2
            Added support for multi-metric shapelet transform
    metric_params : dict, optional
        Parameters for the distance measure. Ignored unless metric is a string.

        Read more about the parameters in the `User guide
        <list_of_subsequence_metrics>`__.
    criterion : {"entropy", "gini"}, optional
        The criterion used to evaluate the utility of a split.
    oob_score : bool, optional
        Use out-of-bag samples to estimate generalization performance. Requires
        `bootstrap=True`.
    bootstrap : bool, optional
        If the samples are drawn with replacement.
    warm_start : bool, optional
        When set to `True`, reuse the solution of the previous call to
        fit and add more estimators to the ensemble, otherwise, just fit a
        whole new ensemble.
    class_weight : dict or "balanced", optional
        Weights associated with the labels

        - if `dict`, weights on the form `{label: weight}`.
        - if "balanced" each class weight inversely proportional to
          the class frequency.
        - if `None`, each class has equal weight.
    n_jobs : int, optional
        The number of jobs to run in parallel. A value of `None` means
        using a single core and a value of `-1` means using all cores.
        Positive integers mean the exact number of cores.
    random_state : int or RandomState, optional
        Controls the random resampling of the original dataset.

        - If `int`, `random_state` is the seed used by the
          random number generator
        - If :class:`numpy.random.RandomState` instance, `random_state` is
          the random number generator
        - If `None`, the random number generator is the
          :class:`numpy.random.RandomState` instance used by
          :func:`numpy.random`.

    Examples
    --------
    >>> from wildboar.ensemble import ShapeletForestClassifier
    >>> from wildboar.datasets import load_synthetic_control
    >>> x, y = load_synthetic_control()
    >>> f = ShapeletForestClassifier(n_estimators=100, metric='scaled_euclidean')
    >>> f.fit(x, y)
    ShapeletForestClassifier(metric='scaled_euclidean')
    >>> y_hat = f.predict(x)
    """

    _parameter_constraints: dict = {
        **ForestMixin._parameter_constraints,
        **ShapeletTreeClassifier._parameter_constraints,
    }

    def __init__(  # noqa: PLR0913 # noqa: PLR0913
        self,
        n_estimators=100,
        *,
        n_shapelets="log2",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        min_shapelet_size=0.0,
        max_shapelet_size=1.0,
        alpha=None,
        metric="euclidean",
        metric_params=None,
        criterion="entropy",
        oob_score=False,
        bootstrap=True,
        warm_start=False,
        class_weight=None,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(
            estimator=ShapeletTreeClassifier(),
            estimator_params=(
                "max_depth",
                "n_shapelets",
                "min_samples_split",
                "min_samples_leaf",
                "min_impurity_decrease",
                "min_shapelet_size",
                "max_shapelet_size",
                "alpha",
                "metric",
                "metric_params",
                "criterion",
            ),
            n_shapelets=n_shapelets,
            n_estimators=n_estimators,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
            min_shapelet_size=min_shapelet_size,
            max_shapelet_size=max_shapelet_size,
            metric=metric,
            metric_params=metric_params,
            criterion=criterion,
            oob_score=oob_score,
            bootstrap=bootstrap,
            warm_start=warm_start,
            class_weight=class_weight,
            n_jobs=n_jobs,
            random_state=random_state,
        )
        self.alpha = alpha


class ExtraShapeletTreesClassifier(BaseShapeletForestClassifier):
    """
    An ensemble of extremely random shapelet trees.

    Parameters
    ----------
    n_estimators : int, optional
        The number of estimators.
    max_depth : int, optional
        The maximum depth of the tree. If `None` the tree is expanded
        until all leaves are pure or until all leaves contain less than
        `min_samples_split` samples.
    min_samples_split : int, optional
        The minimum number of samples to split an internal node.
    min_samples_leaf : int, optional
        The minimum number of samples in a leaf.
    min_impurity_decrease : float, optional
        A split will be introduced only if the impurity decrease is larger than
        or equal to this value.
    min_shapelet_size : float, optional
        The minimum length of a shapelets expressed as a fraction of
        *n_timestep*.
    max_shapelet_size : float, optional
        The maximum length of a shapelets expressed as a fraction of
        *n_timestep*.
    metric : str or list, optional
        The distance metric.

        - If `str`, the distance metric used to identify the best
          shapelet.
        - If `list`, multiple metrics specified as a list of tuples,
          where the first element of the tuple is a metric name and the second
          element a dictionary with a parameter grid specification. A parameter
          grid specification is a dict with two mandatory and one optional
          key-value pairs defining the lower and upper bound on the values and
          number of values in the grid. For example, to specifiy a grid over
          the argument `r` with 10 values in the range 0 to 1, we would give
          the following specification: `dict(min_r=0, max_r=1, num_r=10)`.

          Read more about metric specifications in the `User guide
          <metric_specification>`__.

        .. versionchanged:: 1.2
            Added support for multi-metric shapelet transform
    metric_params : dict, optional
        Parameters for the distance measure. Ignored unless metric is a string.

        Read more about the parameters in the `User guide
        <list_of_subsequence_metrics>`__.
    criterion : {"entropy", "gini"}, optional
        The criterion used to evaluate the utility of a split.
    oob_score : bool, optional
        Use out-of-bag samples to estimate generalization performance. Requires
        `bootstrap=True`.
    bootstrap : bool, optional
        If the samples are drawn with replacement.
    warm_start : bool, optional
        When set to `True`, reuse the solution of the previous call to
        fit and add more estimators to the ensemble, otherwise, just fit a
        whole new ensemble.
    class_weight : dict or "balanced", optional
        Weights associated with the labels

        - if `dict`, weights on the form `{label: weight}`.
        - if "balanced" each class weight inversely proportional to
          the class frequency.
        - if :class:`None`, each class has equal weight.
    n_jobs : int, optional
        The number of jobs to run in parallel. A value of `None` means
        using a single core and a value of `-1` means using all cores.
        Positive integers mean the exact number of cores.
    random_state : int or RandomState, optional
        Controls the random resampling of the original dataset.

        - If `int`, `random_state` is the seed used by the
          random number generator
        - If :class:`numpy.random.RandomState` instance, `random_state` is
          the random number generator
        - If `None`, the random number generator is the
          :class:`numpy.random.RandomState` instance used by
          :func:`numpy.random`.

    Examples
    --------
    >>> from wildboar.ensemble import ExtraShapeletTreesClassifier
    >>> from wildboar.datasets import load_synthetic_control
    >>> x, y = load_synthetic_control()
    >>> f = ExtraShapeletTreesClassifier(n_estimators=100, metric='scaled_euclidean')
    >>> f.fit(x, y)
    ExtraShapeletTreesClassifier(metric='scaled_euclidean')
    >>> y_hat = f.predict(x)
    """

    _parameter_constraints: dict = {
        **ForestMixin._parameter_constraints,
        **ExtraShapeletTreeClassifier._parameter_constraints,
    }
    _parameter_constraints.pop("n_shapelets")

    def __init__(  # noqa: PLR0913 # noqa: PLR0913
        self,
        n_estimators=100,
        *,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        min_shapelet_size=0.0,
        max_shapelet_size=1.0,
        metric="euclidean",
        metric_params=None,
        criterion="entropy",
        oob_score=False,
        bootstrap=True,
        warm_start=False,
        class_weight=None,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(
            estimator=ExtraShapeletTreeClassifier(),
            estimator_params=(
                "max_depth",
                "min_impurity_decrease",
                "min_samples_leaf",
                "min_samples_split",
                "min_shapelet_size",
                "max_shapelet_size",
                "metric",
                "metric_params",
                "criterion",
            ),
            n_shapelets=1,
            n_estimators=n_estimators,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
            min_shapelet_size=min_shapelet_size,
            max_shapelet_size=max_shapelet_size,
            metric=metric,
            metric_params=metric_params,
            criterion=criterion,
            oob_score=oob_score,
            bootstrap=bootstrap,
            warm_start=warm_start,
            class_weight=class_weight,
            n_jobs=n_jobs,
            random_state=random_state,
        )


class BaggingRegressor(BaseBagging, SklearnBaggingRegressor):
    """
    A bagging regressor.

    A bagging regressor is a meta-estimator that fits base classifiers
    on random subsets of the original data.

    Parameters
    ----------
    estimator : object, optional
        Base estimator of the ensemble. If `None`, the base estimator
        is a :class:`~wildboar.tree.ShapeletTreeRegressor`.
    n_estimators : int, optional
        The number of base estimators in the ensemble.
    max_samples : int or float, optional
        The number of samples to draw from `X` to train each base estimator.

        - if `int`, then draw `max_samples` samples.
        - if `float`, then draw `max_samples * n_samples` samples.
    bootstrap : bool, optional
        If the samples are drawn with replacement.
    oob_score : bool, optional
        Use out-of-bag samples to estimate generalization performance. Requires
        `bootstrap=True`.
    warm_start : bool, optional
        When set to `True`, reuse the solution of the previous call
        to fit and add more estimators to the ensemble, otherwise, just fit
        a whole new ensemble.
    n_jobs : int, optional
        The number of jobs to run in parallel. A value of `None` means
        using a single core and a value of `-1` means using all cores.
        Positive integers mean the exact number of cores.
    random_state : int or RandomState, optional
        Controls the random resampling of the original dataset.

        - If `int`, `random_state` is the seed used by the random
          number generator.
        - If :class:`numpy.random.RandomState` instance, `random_state` is
          the random number generator.
        - If `None`, the random number generator is the
          :class:`numpy.random.RandomState` instance used by
          :func:`numpy.random`.
    verbose : int, optional
        Controls the output to standard error while fitting and predicting.
    base_estimator : object, optional
        Use `estimator` instead.

        .. deprecated:: 1.2
            `base_estimator` has been deprecated and will be removed in 1.4.
            Use `estimator` instead.
    """

    _parameter_constraints: dict = {**BaseBagging._parameter_constraints}

    def __init__(
        self,
        estimator=None,
        n_estimators=100,
        *,
        max_samples=1.0,
        bootstrap=True,
        oob_score=False,
        warm_start=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        base_estimator="deprecated",
    ):
        super().__init__(
            estimator=ShapeletTreeRegressor() if estimator is None else estimator,
            n_estimators=n_estimators,
            max_samples=max_samples,
            bootstrap=bootstrap,
            oob_score=oob_score,
            warm_start=warm_start,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            base_estimator=base_estimator,
        )

    def fit(self, x, y, sample_weight=None):
        x, y = self._validate_data(x, y, allow_3d=True, dtype=float, y_numeric=True)

        super()._fit(x, y, self.max_samples, self.max_depth, sample_weight)
        return self


class BaseForestRegressor(ForestMixin, BaggingRegressor, metaclass=ABCMeta):
    @abstractmethod
    def __init__(
        self,
        *,
        estimator,
        estimator_params=tuple(),
        oob_score=False,
        n_estimators=100,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        criterion="squared_error",
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(
            estimator=estimator,
            n_estimators=n_estimators,
            bootstrap=bootstrap,
            n_jobs=n_jobs,
            warm_start=warm_start,
            random_state=random_state,
            oob_score=oob_score,
        )
        self.estimator_params = estimator_params
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.criterion = criterion
        self.min_samples_leaf = min_samples_leaf
        self.min_impurity_decrease = min_impurity_decrease

    def _fit(
        self,
        X,
        y,
        max_samples=None,
        max_depth=None,
        sample_weight=None,
        check_input=False,
    ):
        return super()._fit(X, y, max_samples, max_depth, sample_weight, False)

    def _parallel_args(self):
        return {"prefer": "threads"}


class BaseShapeletForestRegressor(BaseForestRegressor):
    """
    Base class for shapelet forest classifiers.

    Warnings
    --------
    This class should not be used directly. Use derived classes
    instead.
    """

    def __init__(  # noqa: PLR0913
        self,
        *,
        estimator,
        estimator_params=tuple(),
        oob_score=False,
        n_estimators=100,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        n_shapelets="log2",
        min_shapelet_size=0.0,
        max_shapelet_size=1.0,
        metric="euclidean",
        metric_params=None,
        criterion="squared_error",
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(
            estimator=estimator,
            estimator_params=estimator_params,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
            criterion=criterion,
            n_estimators=n_estimators,
            bootstrap=bootstrap,
            n_jobs=n_jobs,
            warm_start=warm_start,
            random_state=random_state,
            oob_score=oob_score,
        )
        self.n_shapelets = n_shapelets
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.metric = metric
        self.metric_params = metric_params

    def _parallel_args(self):
        return {"prefer": "threads"}


class ShapeletForestRegressor(BaseShapeletForestRegressor):
    """
    An ensemble of random shapelet tree regressors.

    Parameters
    ----------
    n_estimators : int, optional
        The number of estimators.
    n_shapelets : int, optional
        The number of shapelets to sample at each node.
    max_depth : int, optional
        The maximum depth of the tree. If `None` the tree is
        expanded until all leaves are pure or until all leaves contain less
        than `min_samples_split` samples.
    min_samples_split : int, optional
        The minimum number of samples to split an internal node.
    min_samples_leaf : int, optional
        The minimum number of samples in a leaf.
    min_impurity_decrease : float, optional
        A split will be introduced only if the impurity decrease is larger
        than or equal to this value.
    min_shapelet_size : float, optional
        The minimum length of a shapelets expressed as a fraction of
        *n_timestep*.
    max_shapelet_size : float, optional
        The maximum length of a shapelets expressed as a fraction of
        *n_timestep*.
    alpha : float, optional
        Dynamically decrease the number of sampled shapelets at each node
        according to the current depth, i.e.:::
            w = 1 - exp(-abs(alpha) * depth)

        - if `alpha < 0`, the number of sampled shapelets decrease from
          `n_shapelets` towards 1 with increased depth.
        - if `alpha > 0`, the number of sampled shapelets increase from `1`
          towards `n_shapelets` with increased depth.
        - if `None`, the number of sampled shapelets are the same
          independeth of depth.
    metric : str or list, optional
        The distance metric.

        - If `str`, the distance metric used to identify the best
          shapelet.
        - If `list`, multiple metrics specified as a list of tuples,
          where the first element of the tuple is a metric name and the second
          element a dictionary with a parameter grid specification. A parameter
          grid specification is a dict with two mandatory and one optional
          key-value pairs defining the lower and upper bound on the values and
          number of values in the grid. For example, to specifiy a grid over
          the argument `r` with 10 values in the range 0 to 1, we would give
          the following specification: `dict(min_r=0, max_r=1, num_r=10)`.

          Read more about metric specifications in the `User guide
          <metric_specification>`__.

        .. versionchanged:: 1.2
            Added support for multi-metric shapelet transform
    metric_params : dict, optional
        Parameters for the distance measure. Ignored unless metric is a string.

        Read more about the parameters in the `User guide
        <list_of_subsequence_metrics>`__.
    criterion : {"squared_error"}, optional
        The criterion used to evaluate the utility of a split.

        .. deprecated:: 1.1
            Criterion "mse" was deprecated in v1.1 and removed in version 1.2.
    oob_score : bool, optional
        Use out-of-bag samples to estimate generalization performance. Requires
        `bootstrap=True`.
    bootstrap : bool, optional
        If the samples are drawn with replacement.
    warm_start : bool, optional
        When set to `True`, reuse the solution of the previous call to
        fit and add more estimators to the ensemble, otherwise, just fit a
        whole new ensemble.
    n_jobs : int, optional
        The number of processor cores used for fitting the ensemble.
    random_state : int or RandomState, optional
        Controls the random resampling of the original dataset.

        - If `int`, `random_state` is the seed used by the
          random number generator
        - If :class:`numpy.random.RandomState` instance, `random_state` is
          the random number generator
        - If `None`, the random number generator is the
          :class:`numpy.random.RandomState` instance used by
          :func:`numpy.random`.

    Examples
    --------
    >>> from wildboar.ensemble import ShapeletForestRegressor
    >>> from wildboar.datasets import load_synthetic_control
    >>> x, y = load_synthetic_control()
    >>> f = ShapeletForestRegressor(n_estimators=100, metric='scaled_euclidean')
    >>> f.fit(x, y)
    ShapeletForestRegressor(metric='scaled_euclidean')
    >>> y_hat = f.predict(x)
    """

    _parameter_constraints: dict = {
        **ForestMixin._parameter_constraints,
        **ShapeletTreeRegressor._parameter_constraints,
    }

    def __init__(  # noqa: PLR0913
        self,
        n_estimators=100,
        *,
        n_shapelets="log2",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        min_shapelet_size=0.0,
        max_shapelet_size=1.0,
        alpha=None,
        metric="euclidean",
        metric_params=None,
        criterion="squared_error",
        oob_score=False,
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(
            estimator=ShapeletTreeRegressor(),
            estimator_params=(
                "max_depth",
                "n_shapelets",
                "min_samples_split",
                "min_samples_leaf",
                "min_impurity_decrease",
                "min_shapelet_size",
                "max_shapelet_size",
                "alpha",
                "metric",
                "metric_params",
                "criterion",
            ),
            n_shapelets=n_shapelets,
            n_estimators=n_estimators,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
            min_shapelet_size=min_shapelet_size,
            max_shapelet_size=max_shapelet_size,
            metric=metric,
            metric_params=metric_params,
            criterion=criterion,
            oob_score=oob_score,
            bootstrap=bootstrap,
            warm_start=warm_start,
            n_jobs=n_jobs,
            random_state=random_state,
        )
        self.alpha = alpha


class ExtraShapeletTreesRegressor(BaseShapeletForestRegressor):
    """
    An ensemble of extremely random shapelet tree regressors.

    Parameters
    ----------
    n_estimators : int, optional
        The number of estimators.
    max_depth : int, optional
        The maximum depth of the tree. If `None` the tree is expanded
        until all leaves are pure or until all leaves contain less than
        `min_samples_split` samples.
    min_samples_split : int, optional
        The minimum number of samples to split an internal node.
    min_shapelet_size : float, optional
        The minimum length of a shapelets expressed as a fraction of
        *n_timestep*.
    max_shapelet_size : float, optional
        The maximum length of a shapelets expressed as a fraction of
        *n_timestep*.
    metric : str or list, optional
        The distance metric.

        - If `str`, the distance metric used to identify the best
          shapelet.
        - If `list`, multiple metrics specified as a list of tuples,
          where the first element of the tuple is a metric name and the second
          element a dictionary with a parameter grid specification. A parameter
          grid specification is a dict with two mandatory and one optional
          key-value pairs defining the lower and upper bound on the values and
          number of values in the grid. For example, to specifiy a grid over
          the argument `r` with 10 values in the range 0 to 1, we would give
          the following specification: `dict(min_r=0, max_r=1, num_r=10)`.

          Read more about metric specifications in the `User guide
          <metric_specification>`__.

        .. versionchanged:: 1.2
            Added support for multi-metric shapelet transform
    metric_params : dict, optional
        Parameters for the distance measure. Ignored unless metric is a string.

        Read more about the parameters in the `User guide
        <list_of_subsequence_metrics>`__.
    criterion : {"squared_error"}, optional
        The criterion used to evaluate the utility of a split.

        .. deprecated:: 1.1
            Criterion "mse" was deprecated in v1.1 and removed in version 1.2.
    oob_score : bool, optional
        Use out-of-bag samples to estimate generalization performance. Requires
        `bootstrap=True`.
    bootstrap : bool, optional
        If the samples are drawn with replacement.
    warm_start : bool, optional
        When set to `True`, reuse the solution of the previous call to
        fit and add more estimators to the ensemble, otherwise, just fit a
        whole new ensemble.
    n_jobs : int, optional
        The number of jobs to run in parallel. A value of `None` means
        using a single core and a value of `-1` means using all cores.
        Positive integers mean the exact number of cores.
    random_state : int or RandomState, optional
        Controls the random resampling of the original dataset.

        - If int, `random_state` is the seed used by the
          random number generator
        - If :class:`numpy.random.RandomState` instance, `random_state` is
          the random number generator
        - If `None`, the random number generator is the
          :class:`numpy.random.RandomState` instance used by
          :func:`numpy.random`.

    Examples
    --------
    >>> from wildboar.ensemble import ExtraShapeletTreesRegressor
    >>> from wildboar.datasets import load_synthetic_control
    >>> x, y = load_synthetic_control()
    >>> f = ExtraShapeletTreesRegressor(n_estimators=100, metric='scaled_euclidean')
    >>> f.fit(x, y)
    ExtraShapeletTreesRegressor(metric='scaled_euclidean')
    >>> y_hat = f.predict(x)
    """

    _parameter_constraints: dict = {
        **ForestMixin._parameter_constraints,
        **ExtraShapeletTreeRegressor._parameter_constraints,
    }
    _parameter_constraints.pop("n_shapelets")

    def __init__(
        self,
        n_estimators=100,
        *,
        max_depth=None,
        min_samples_split=2,
        min_shapelet_size=0,
        max_shapelet_size=1,
        metric="euclidean",
        metric_params=None,
        criterion="squared_error",
        oob_score=False,
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(
            estimator=ExtraShapeletTreeRegressor(),
            estimator_params=(
                "max_depth",
                "min_impurity_decrease",
                "min_samples_leaf",
                "min_samples_split",
                "min_shapelet_size",
                "max_shapelet_size",
                "metric",
                "metric_params",
                "criterion",
            ),
            n_shapelets=1,
            n_estimators=n_estimators,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_shapelet_size=min_shapelet_size,
            max_shapelet_size=max_shapelet_size,
            metric=metric,
            metric_params=metric_params,
            criterion=criterion,
            oob_score=oob_score,
            bootstrap=bootstrap,
            warm_start=warm_start,
            n_jobs=n_jobs,
            random_state=random_state,
        )


class ShapeletForestEmbedding(BaseShapeletForestRegressor):
    """
    An ensemble of random shapelet trees.

    An unsupervised transformation of a time series dataset to a
    high-dimensional sparse representation. A time series i indexed by the leaf
    that it falls into. This leads to a binary coding of a time series with as
    many ones as trees in the forest.

    The dimensionality of the resulting representation is `<= n_estimators *
    2^max_depth`

    Parameters
    ----------
    n_estimators : int, optional
        The number of estimators.
    n_shapelets : int, optional
        The number of shapelets to sample at each node.
    max_depth : int, optional
        The maximum depth of the tree. If `None` the tree is expanded
        until all leaves are pure or until all leaves contain less than
        `min_samples_split` samples.
    min_samples_split : int, optional
        The minimum number of samples to split an internal node.
    min_samples_leaf : int, optional
        The minimum number of samples in a leaf.
    min_impurity_decrease : float, optional
        A split will be introduced only if the impurity decrease is larger than
        or equal to this value.
    min_shapelet_size : float, optional
        The minimum length of a shapelets expressed as a fraction of
        *n_timestep*.
    max_shapelet_size : float, optional
        The maximum length of a shapelets expressed as a fraction of
        *n_timestep*.
    metric : str or list, optional
        The distance metric.

        - If `str`, the distance metric used to identify the best
          shapelet.
        - If `list`, multiple metrics specified as a list of tuples,
          where the first element of the tuple is a metric name and the second
          element a dictionary with a parameter grid specification. A parameter
          grid specification is a dict with two mandatory and one optional
          key-value pairs defining the lower and upper bound on the values and
          number of values in the grid. For example, to specifiy a grid over
          the argument `r` with 10 values in the range 0 to 1, we would give
          the following specification: `dict(min_r=0, max_r=1, num_r=10)`.

          Read more about metric specifications in the `User guide
          <metric_specification>`__.

        .. versionchanged:: 1.2
            Added support for multi-metric shapelet transform
    metric_params : dict, optional
        Parameters for the distance measure. Ignored unless metric is a string.

        Read more about the parameters in the `User guide
        <list_of_subsequence_metrics>`__.
    criterion : {"squared_error"}, optional
        The criterion used to evaluate the utility of a split.

        .. deprecated:: 1.1
            Criterion "mse" was deprecated in v1.1 and removed in version 1.2.
    bootstrap : bool, optional
        If the samples are drawn with replacement.
    warm_start : bool, optional
        When set to `True`, reuse the solution of the previous call to
        fit and add more estimators to the ensemble, otherwise, just fit a
        whole new ensemble.
    n_jobs : int, optional
        The number of jobs to run in parallel. A value of `None` means
        using a single core and a value of `-1` means using all cores.
        Positive integers mean the exact number of cores.
    sparse_output : bool, optional
        Return a sparse CSR-matrix.
    random_state : int or RandomState, optional
        Controls the random resampling of the original dataset.

        - If `int`, `random_state` is the seed used by the
          random number generator.
        - If :class:`numpy.random.RandomState` instance, `random_state` is
          the random number generator.
        - If `None`, the random number generator is the
          :class:`numpy.random.RandomState` instance used by
          :func:`numpy.random`.
    """

    _parameter_constraints: dict = {
        **ForestMixin._parameter_constraints,
        **ExtraShapeletTreeRegressor._parameter_constraints,
        "sparse_output": ["boolean"],
    }

    def __init__(  # noqa: PLR0913
        self,
        n_estimators=100,
        *,
        n_shapelets=1,
        max_depth=5,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        min_shapelet_size=0.0,
        max_shapelet_size=1.0,
        metric="euclidean",
        metric_params=None,
        criterion="squared_error",
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        sparse_output=True,
        random_state=None,
    ):
        super().__init__(
            estimator=ExtraShapeletTreeRegressor(),
            estimator_params=(
                "n_shapelets",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "min_impurity_decrease",
                "min_shapelet_size",
                "max_shapelet_size",
                "metric",
                "metric_params",
                "criterion",
            ),
            n_shapelets=n_shapelets,
            n_estimators=n_estimators,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
            min_shapelet_size=min_shapelet_size,
            max_shapelet_size=max_shapelet_size,
            metric=metric,
            metric_params=metric_params,
            criterion=criterion,
            oob_score=False,
            bootstrap=bootstrap,
            warm_start=warm_start,
            n_jobs=n_jobs,
            random_state=random_state,
        )
        self.sparse_output = sparse_output

    def _set_oob_score(self, X, y):
        raise NotImplementedError("OOB score not supported")

    def fit(self, x, y=None, sample_weight=None):
        self.fit_transform(x, y, sample_weight)
        return self

    def fit_transform(self, x, y=None, sample_weight=None):
        x = self._validate_data(x, allow_3d=True, dtype=float)
        random_state = check_random_state(self.random_state)
        y = random_state.uniform(size=x.shape[0])
        super().fit(x, y, sample_weight=sample_weight)
        self.one_hot_encoder_ = OneHotEncoder(
            sparse=self.sparse_output, handle_unknown="ignore"
        )
        return self.one_hot_encoder_.fit_transform(self.apply(x))

    def transform(self, x):
        check_is_fitted(self)
        return self.one_hot_encoder_.transform(self.apply(x))


class IsolationShapeletForest(OutlierMixin, ForestMixin, BaseBagging):
    """
    An isolation shapelet forest.

    .. versionadded:: 0.3.5

    Parameters
    ----------
    n_estimators : int, optional
        The number of estimators in the ensemble.
    n_shapelets : int, optional
        The number of shapelets to sample at each node.
    bootstrap : bool, optional
        If the samples are drawn with replacement.
    n_jobs : int, optional
        The number of jobs to run in parallel. A value of `None` means using a
        single core and a value of `-1` means using all cores. Positive
        integers mean the exact number of cores.
    min_shapelet_size : float, optional
        The minimum length of a shapelets expressed as a fraction of
        *n_timestep*.
    max_shapelet_size : float, optional
        The maximum length of a shapelets expressed as a fraction of
        *n_timestep*.
    min_samples_split : int, optional
        The minimum number of samples to split an internal node.
    max_samples : "auto", float or int, optional
        The number of samples to draw to train each base estimator.
    contamination : 'auto' or float, optional
        The strategy for computing the offset.

        - if "auto" then `offset_` is set to `-0.5`.
        - if float `offset_` is computed as the c:th percentile of
          scores.

        If `bootstrap=True`, out-of-bag samples are used for computing
        the scores.
    warm_start : bool, optional
        When set to `True`, reuse the solution of the previous call to
        fit and add more estimators to the ensemble, otherwise, just fit a
        whole new ensemble.
    metric : str or list, optional
        The distance metric.

        - If `str`, the distance metric used to identify the best
          shapelet.
        - If `list`, multiple metrics specified as a list of tuples,
          where the first element of the tuple is a metric name and the second
          element a dictionary with a parameter grid specification. A parameter
          grid specification is a dict with two mandatory and one optional
          key-value pairs defining the lower and upper bound on the values and
          number of values in the grid. For example, to specifiy a grid over
          the argument `r` with 10 values in the range 0 to 1, we would give
          the following specification: `dict(min_r=0, max_r=1, num_r=10)`.

          Read more about metric specifications in the `User guide
          <metric_specification>`__.

        .. versionchanged:: 1.2
            Added support for multi-metric shapelet transform
    metric_params : dict, optional
        Parameters for the distance measure. Ignored unless metric is a string.

        Read more about the parameters in the `User guide
        <list_of_subsequence_metrics>`__.
    random_state : int or RandomState, optional
        Controls the random resampling of the original dataset.

        - If `int`, `random_state` is the seed used by the
          random number generator.
        - If :class:`numpy.random.RandomState` instance, `random_state` is
          the random number generator.
        - If `None`, the random number generator is the
          :class:`numpy.random.RandomState` instance used by
          :func:`numpy.random`.

    Attributes
    ----------
    offset_ : float
        The offset for computing the final decision

    Examples
    --------
    Using default offset threshold

    >>> from wildboar.ensemble import IsolationShapeletForest
    >>> from wildboar.datasets import load_two_lead_ecg
    >>> from wildboar.model_selection import outlier_train_test_split
    >>> from sklearn.metrics import balanced_accuracy_score
    >>> f = IsolationShapeletForest(random_state=1)
    >>> x, y = load_two_lead_ecg()
    >>> x_train, x_test, y_train, y_test = outlier_train_test_split(
    ...     x, y, 1, test_size=0.2, anomalies_train_size=0.05, random_state=1
    ... )
    >>> f.fit(x_train)
    IsolationShapeletForest(random_state=1)
    >>> y_pred = f.predict(x_test)
    >>> balanced_accuracy_score(y_test, y_pred) # doctest: +NUMBER
    0.8674
    """

    _parameter_constraints: dict = {
        **ForestMixin._parameter_constraints,
        **ShapeletMixin._parameter_constraints,
        "contamination": [StrOptions({"auto"}), Interval(Real, 0, 0.5, closed="right")],
        "max_samples": [
            None,
            Interval(Real, 0.0, 1.0, closed="right"),
            Interval(Integral, 1, None, closed="left"),
            StrOptions({"auto"}),
        ],
    }

    def __init__(
        self,
        n_estimators=100,
        *,
        n_shapelets=1,
        bootstrap=False,
        n_jobs=None,
        min_shapelet_size=0,
        max_shapelet_size=1,
        min_samples_split=2,
        max_samples="auto",
        contamination="auto",
        warm_start=False,
        metric="euclidean",
        metric_params=None,
        random_state=None,
    ):
        super(IsolationShapeletForest, self).__init__(
            estimator=ExtraShapeletTreeRegressor(),
            bootstrap=bootstrap,
            warm_start=warm_start,
            n_jobs=n_jobs,
            n_estimators=n_estimators,
            random_state=random_state,
        )
        self.estimator_params = (
            "n_shapelets",
            "min_samples_split",
            "min_shapelet_size",
            "max_shapelet_size",
            "metric",
            "metric_params",
        )
        self.n_shapelets = n_shapelets
        self.metric = metric
        self.metric_params = metric_params
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.min_samples_split = min_samples_split
        self.contamination = contamination
        self.max_samples = max_samples

    def _set_oob_score(self, x, y):
        raise NotImplementedError("OOB score not supported")

    def _parallel_args(self):
        return {"prefer": "threads"}

    def fit(self, x, y=None, sample_weight=None):
        x = self._validate_data(x, allow_3d=True, dtype=float)
        random_state = check_random_state(self.random_state)

        rnd_y = random_state.uniform(size=x.shape[0])
        if isinstance(self.max_samples, str):
            if self.max_samples == "auto":
                max_samples = min(x.shape[0], 256)
            else:
                raise ValueError(
                    "max_samples must be 'auto', int or float, got %r."
                    % self.max_samples
                )
        elif isinstance(self.max_samples, numbers.Integral):
            max_samples = min(self.max_samples, x.shape[0])
        else:
            max_samples = math.ceil(
                check_scalar(
                    self.max_samples,
                    "max_samples",
                    numbers.Real,
                    min_val=0,
                    max_val=1.0,
                    include_boundaries="right",
                )
                * x.shape[0]
            )

        max_depth = int(np.ceil(np.log2(max(max_samples, 2))))
        super(IsolationShapeletForest, self)._fit(
            x,
            rnd_y,
            max_samples=max_samples,
            max_depth=max_depth,
            sample_weight=sample_weight,
            check_input=False,
        )

        self.max_samples_ = max_samples
        if self.contamination == "auto":
            self.offset_ = -0.5
        elif isinstance(self.contamination, numbers.Real):
            if not 0 < self.contamination <= 0.5:
                raise ValueError("contamination must be in (0, 0.5]")
            if self.bootstrap:
                scores = self._oob_score_samples(x)
            else:
                scores = self.score_samples(x)

            self.offset_ = np.percentile(scores, 100.0 * self.contamination)
        else:
            raise ValueError(
                "contamination must be 'auto' or float, got %r." % self.contamination
            )

        return self

    def predict(self, x):
        decision = self.decision_function(x)
        is_inlier = np.ones(decision.shape[0], dtype=int)
        is_inlier[decision < 0] = -1
        return is_inlier

    def decision_function(self, x):
        return self.score_samples(x) - self.offset_

    def score_samples(self, x):
        check_is_fitted(self)
        x = self._validate_data(x, reset=False, allow_3d=True, dtype=float)
        return _score_samples(x, self.estimators_, self.max_samples_)

    def _oob_score_samples(self, x):
        n_samples = x.shape[0]
        score_samples = np.zeros((n_samples,))

        # We need to make this more efficient
        for i in range(x.shape[0]):
            estimators = []
            for estimator, samples in zip(self.estimators_, self.estimators_samples_):
                if i not in samples:
                    estimators.append(estimator)
            score_samples[i] = _score_samples(x[[i]], estimators, self.max_samples_)
        return score_samples

    def _more_tags(self):
        return {
            "_xfail_checks": {
                "check_sample_weights_invariance": (
                    "zero sample_weight is not equivalent to removing samples"
                ),
            }
        }


def _score_samples(x, estimators, max_samples):
    # From: https://github.com/scikit-learn/scikit-learn/blob/
    # 0fb307bf39bbdacd6ed713c00724f8f871d60370/sklearn/ensemble/_iforest.py#L411
    depths = np.zeros(x.shape[0], order="F")

    for tree in estimators:
        leaves_index = tree.apply(x, check_input=False)
        node_indicator = tree.decision_path(x, check_input=False)
        n_samples_leaf = tree.tree_.n_node_samples[leaves_index]

        depths += (
            np.ravel(node_indicator.sum(axis=1))
            + _average_path_length(n_samples_leaf)
            - 1.0
        )

    denominator = len(estimators) * _average_path_length(np.array([max_samples]))
    scores = 2 ** -np.divide(
        depths, denominator, out=np.ones_like(depths), where=denominator != 0
    )
    return -scores


def _average_path_length(n_samples_leaf):
    # From: https://github.com/scikit-learn/scikit-learn/blob/
    # 0fb307bf39bbdacd6ed713c00724f8f871d60370/sklearn/ensemble/_iforest.py#L480
    n_samples_leaf_shape = n_samples_leaf.shape
    n_samples_leaf = n_samples_leaf.reshape((1, -1))
    average_path_length = np.zeros(n_samples_leaf.shape)

    mask_1 = n_samples_leaf <= 1
    mask_2 = n_samples_leaf == 2
    not_mask = ~np.logical_or(mask_1, mask_2)

    average_path_length[mask_1] = 0.0
    average_path_length[mask_2] = 1.0
    average_path_length[not_mask] = (
        2.0 * (np.log(n_samples_leaf[not_mask] - 1.0) + np.euler_gamma)
        - 2.0 * (n_samples_leaf[not_mask] - 1.0) / n_samples_leaf[not_mask]
    )

    return average_path_length.reshape(n_samples_leaf_shape)


class RocketForestRegressor(BaseForestRegressor):
    """
    An ensemble of rocket tree regressors.

    Parameters
    ----------
    n_estimators : int, optional
        The number of estimators.
    n_kernels : int, optional
        The number of shapelets to sample at each node.
    oob_score : bool, optional
        Use out-of-bag samples to estimate generalization performance. Requires
        `bootstrap=True`.
    max_depth : int, optional
        The maximum depth of the tree. If `None` the tree is
        expanded until all leaves are pure or until all leaves contain less
        than `min_samples_split` samples.
    min_samples_split : int, optional
        The minimum number of samples to split an internal node.
    min_samples_leaf : int, optional
        The minimum number of samples in a leaf.
    min_impurity_decrease : float, optional
        A split will be introduced only if the impurity decrease is larger
        than or equal to this value.
    sampling : {"normal", "uniform", "shapelet"}, optional
        The sampling of convolutional filters.

        - if "normal", sample filter according to a normal distribution with
            ``mean`` and ``scale``.
        - if "uniform", sample filter according to a uniform distribution with
            ``lower`` and ``upper``.
        - if "shapelet", sample filters as subsequences in the training data.
    sampling_params : dict, optional
        The parameters for the sampling.

        - if "normal", ``{"mean": float, "scale": float}``, defaults to
            ``{"mean": 0, "scale": 1}``.
        - if "uniform", ``{"lower": float, "upper": float}``, defaults to
            ``{"lower": -1, "upper": 1}``.
    kernel_size : array-like, optional
        The kernel size, by default ``[7, 11, 13]``.
    min_size : float, optional
        The minimum length of a shapelets expressed as a fraction of
        *n_timestep*.
    max_size : float, optional
        The maximum length of a shapelets expressed as a fraction of
        *n_timestep*.
    bias_prob : float, optional
        The probability of using a bias term.
    normalize_prob : float, optional
        The probability of performing normalization.
    padding_prob : float, optional
        The probability of padding with zeros.
    criterion : {"squared_error"}, optional
        The criterion used to evaluate the utility of a split.

        .. deprecated:: 1.1
            Criterion "mse" was deprecated in v1.1 and removed in version 1.2.
    bootstrap : bool, optional
        If the samples are drawn with replacement.
    warm_start : bool, optional
        When set to `True`, reuse the solution of the previous call to
        fit and add more estimators to the ensemble, otherwise, just fit a
        whole new ensemble.
    n_jobs : int, optional
        The number of processor cores used for fitting the ensemble.
    random_state : int or RandomState, optional
        Controls the random resampling of the original dataset.

        - If `int`, `random_state` is the seed used by the
          random number generator
        - If :class:`numpy.random.RandomState` instance, `random_state` is
          the random number generator
        - If `None`, the random number generator is the
          :class:`numpy.random.RandomState` instance used by
          :func:`numpy.random`.
    """

    _parameter_constraints: dict = {
        **ForestMixin._parameter_constraints,
        **RocketTreeRegressor._parameter_constraints,
    }

    def __init__(  # noqa: PLR0913
        self,
        n_estimators=100,
        *,
        n_kernels=10,
        oob_score=False,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        sampling="normal",
        sampling_params=None,
        kernel_size=None,
        min_size=None,
        max_size=None,
        bias_prob=1.0,
        normalize_prob=1.0,
        padding_prob=0.5,
        criterion="squared_error",
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(
            estimator=RocketTreeRegressor(),
            estimator_params=(
                "n_kernels",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "min_impurity_decrease",
                "kernel_size",
                "min_size",
                "max_size",
                "sampling",
                "sampling_params",
                "bias_prob",
                "normalize_prob",
                "padding_prob",
                "criterion",
            ),
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
            criterion=criterion,
            n_estimators=n_estimators,
            bootstrap=bootstrap,
            n_jobs=n_jobs,
            warm_start=warm_start,
            random_state=random_state,
            oob_score=oob_score,
        )
        self.n_kernels = n_kernels
        self.sampling = sampling
        self.sampling_params = sampling_params
        self.kernel_size = kernel_size
        self.bias_prob = bias_prob
        self.normalize_prob = normalize_prob
        self.padding_prob = padding_prob
        self.min_size = min_size
        self.max_size = max_size

    def _parallel_args(self):
        return {"prefer": "threads"}


class RocketForestClassifier(BaseForestClassifier):
    """An ensemble of rocket tree classifiers."""

    _parameter_constraints: dict = {
        **ForestMixin._parameter_constraints,
        **RocketTreeClassifier._parameter_constraints,
    }

    def __init__(  # noqa: PLR0913
        self,
        n_estimators=100,
        *,
        n_kernels=10,
        oob_score=False,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        sampling="normal",
        sampling_params=None,
        kernel_size=None,
        min_size=None,
        max_size=None,
        bias_prob=1.0,
        normalize_prob=1.0,
        padding_prob=0.5,
        criterion="entropy",
        bootstrap=True,
        warm_start=False,
        class_weight=None,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(
            estimator=RocketTreeClassifier(),
            estimator_params=(
                "n_kernels",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "min_impurity_decrease",
                "kernel_size",
                "min_size",
                "max_size",
                "sampling",
                "sampling_params",
                "bias_prob",
                "normalize_prob",
                "padding_prob",
                "criterion",
            ),
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
            criterion=criterion,
            n_estimators=n_estimators,
            bootstrap=bootstrap,
            n_jobs=n_jobs,
            warm_start=warm_start,
            random_state=random_state,
            oob_score=oob_score,
            class_weight=class_weight,
        )
        self.n_kernels = n_kernels
        self.sampling = sampling
        self.sampling_params = sampling_params
        self.kernel_size = kernel_size
        self.bias_prob = bias_prob
        self.normalize_prob = normalize_prob
        self.padding_prob = padding_prob
        self.min_size = min_size
        self.max_size = max_size

    def _parallel_args(self):
        return {"prefer": "threads"}


class IntervalForestClassifier(BaseForestClassifier):
    """An ensemble of interval tree classifiers."""

    _parameter_constraints: dict = {
        **ForestMixin._parameter_constraints,
        **IntervalTreeClassifier._parameter_constraints,
    }

    def __init__(  # noqa: PLR0913
        self,
        n_estimators=100,
        *,
        n_intervals="sqrt",
        intervals="fixed",
        summarizer="mean_var_std",
        sample_size=0.5,
        min_size=0.0,
        max_size=1.0,
        oob_score=False,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0,
        criterion="entropy",
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        class_weight=None,
        random_state=None,
    ):
        super().__init__(
            estimator=IntervalTreeClassifier(),
            estimator_params=(
                "n_intervals",
                "intervals",
                "summarizer",
                "sample_size",
                "min_size",
                "max_size",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "min_impurity_decrease",
                "criterion",
            ),
            oob_score=oob_score,
            n_estimators=n_estimators,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
            criterion=criterion,
            bootstrap=bootstrap,
            warm_start=warm_start,
            n_jobs=n_jobs,
            class_weight=class_weight,
            random_state=random_state,
        )
        self.n_intervals = n_intervals
        self.intervals = intervals
        self.summarizer = summarizer
        self.sample_size = sample_size
        self.min_size = min_size
        self.max_size = max_size

    def _parallel_args(self):
        return {"prefer": "threads"}


class IntervalForestRegressor(BaseForestRegressor):
    """An ensemble of interval tree regressors."""

    _parameter_constraints: dict = {
        **ForestMixin._parameter_constraints,
        **IntervalTreeRegressor._parameter_constraints,
    }

    def __init__(  # noqa: PLR0913
        self,
        n_estimators=100,
        *,
        n_intervals="sqrt",
        intervals="fixed",
        summarizer="auto",
        sample_size=0.5,
        min_size=0.0,
        max_size=1.0,
        oob_score=False,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0,
        criterion="squared_error",
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(
            estimator=IntervalTreeRegressor(),
            estimator_params=(
                "n_intervals",
                "intervals",
                "summarizer",
                "sample_size",
                "min_size",
                "max_size",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "min_impurity_decrease",
                "criterion",
            ),
            oob_score=oob_score,
            n_estimators=n_estimators,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
            criterion=criterion,
            bootstrap=bootstrap,
            warm_start=warm_start,
            n_jobs=n_jobs,
            random_state=random_state,
        )
        self.n_intervals = n_intervals
        self.intervals = intervals
        self.summarizer = summarizer
        self.sample_size = sample_size
        self.min_size = min_size
        self.max_size = max_size

    def _parallel_args(self):
        return {"prefer": "threads"}


class PivotForestClassifier(BaseForestClassifier):
    """An ensemble of interval tree classifiers."""

    _parameter_constraints: dict = {
        **ForestMixin._parameter_constraints,
        **PivotTreeClassifier._parameter_constraints,
    }

    def __init__(
        self,
        n_estimators=100,
        *,
        n_pivot="sqrt",
        metrics="all",
        oob_score=False,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0,
        criterion="entropy",
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        class_weight=None,
        random_state=None,
    ):
        super().__init__(
            estimator=PivotTreeClassifier(),
            estimator_params=(
                "n_pivot",
                "metrics",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "min_impurity_decrease",
                "criterion",
            ),
            oob_score=oob_score,
            n_estimators=n_estimators,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
            criterion=criterion,
            bootstrap=bootstrap,
            warm_start=warm_start,
            n_jobs=n_jobs,
            class_weight=class_weight,
            random_state=random_state,
        )
        self.n_pivot = n_pivot
        self.metrics = metrics

    def _parallel_args(self):
        return {"prefer": "threads"}


class ProximityForestClassifier(BaseForestClassifier):
    """
    A forest of proximity trees.

    Parameters
    ----------
    n_estimators : int, optional
        The number of estimators.
    n_pivot : int, optional
        The number of pivots to sample at each node.
    pivot_sample : {"label", "uniform"}, optional
        The pivot sampling method.
    metric_sample : {"uniform", "weighted"}, optional
        The metric sampling method.
    metric : {"auto", "default"}, str or list, optional
        The distance metrics. By default, we use the parameterization suggested by
        Lucas et.al (2019).

        - If "auto", use the default metric specification, suggested by
          (Lucas et. al, 2020).
        - If str, use a single metric or default metric specification.
        - If list, custom metric specification can be given as a list of
          tuples, where the first element of the tuple is a metric name and the
          second element a dictionary with a parameter grid specification. A
          parameter grid specification is a `dict` with two mandatory and one
          optional key-value pairs defining the lower and upper bound on the
          values as well as the number of values in the grid. For example, to
          specifiy a grid over the argument 'r' with 10 values in the range 0
          to 1, we would give the following specification:
          `dict(min_r=0, max_r=1, num_r=10)`.

        Read more about the metrics and their parameters in the
        :ref:`User guide <list_of_metrics>`.
    metric_params : dict, optional
        Parameters for the distance measure. Ignored unless metric is a string.

        Read more about the parameters in the :ref:`User guide
        <list_of_metrics>`.
    metric_factories : dict, optional
        A metric specification.

        .. deprecated:: 1.2
            Use the combination of metric and metric params.
    oob_score : bool, optional
        Use out-of-bag samples to estimate generalization performance. Requires
        `bootstrap=True`.
    max_depth : int, optional
        The maximum tree depth.
    min_samples_split : int, optional
        The minimum number of samples to consider a split.
    min_samples_leaf : int, optional
        The minimum number of samples in a leaf.
    min_impurity_decrease : float, optional
        The minimum impurity decrease to build a sub-tree.
    criterion : {"entropy", "gini"}, optional
        The impurity criterion.
    bootstrap : bool, optional
        If the samples are drawn with replacement.
    warm_start : bool, optional
        When set to `True`, reuse the solution of the previous call
        to fit and add more estimators to the ensemble, otherwise, just fit
        a whole new ensemble.
    n_jobs : int, optional
        The number of jobs to run in parallel. A value of `None` means
        using a single core and a value of `-1` means using all cores.
        Positive integers mean the exact number of cores.
    class_weight : dict or "balanced", optional
        Weights associated with the labels.

        - if dict, weights on the form {label: weight}.
        - if "balanced" each class weight inversely proportional to the class
            frequency.
        - if None, each class has equal weight.
    random_state : int or RandomState, optional
        - If `int`, `random_state` is the seed used by the random number generator
        - If `RandomState` instance, `random_state` is the random number generator
        - If `None`, the random number generator is the `RandomState` instance used
            by `np.random`.

    References
    ----------
    Lucas, Benjamin, Ahmed Shifaz, Charlotte Pelletier, Lachlan O'Neill, Nayyar Zaidi, \
    Bart Goethals, François Petitjean, and Geoffrey I. Webb. (2019)
        Proximity forest: an effective and scalable distance-based classifier for time
        series. Data Mining and Knowledge Discovery
    """

    _parameter_constraints: dict = {
        **ForestMixin._parameter_constraints,
        **ProximityTreeClassifier._parameter_constraints,
    }

    def __init__(  # noqa: PLR0913
        self,
        n_estimators=100,
        *,
        n_pivot=1,
        pivot_sample="label",
        metric_sample="weighted",
        metric="auto",
        metric_params=None,
        metric_factories=None,
        oob_score=False,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0,
        criterion="entropy",
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        class_weight=None,
        random_state=None,
    ):
        super().__init__(
            estimator=ProximityTreeClassifier(),
            estimator_params=(
                "n_pivot",
                "metric_factories",
                "metric",
                "metric_params",
                "pivot_sample",
                "metric_sample",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "min_impurity_decrease",
                "criterion",
            ),
            oob_score=oob_score,
            n_estimators=n_estimators,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
            criterion=criterion,
            bootstrap=bootstrap,
            warm_start=warm_start,
            n_jobs=n_jobs,
            class_weight=class_weight,
            random_state=random_state,
        )
        self.n_pivot = n_pivot
        self.metric_factories = metric_factories
        self.metric = metric
        self.metric_params = metric_params
        self.pivot_sample = pivot_sample
        self.metric_sample = metric_sample

    def _parallel_args(self):
        return {"prefer": "threads"}



"""Counterfactual explanations.

The :mod:`wildboar.explain.counterfactual` module includes numerous
methods for generating counterfactual explanations.
"""
# Authors: Isak Samsten
# License: BSD 3 clause

from ._helper import _proximity as proximity
from ._helper import counterfactuals
from ._nn import KNeighborsCounterfactual
from ._proto import PrototypeCounterfactual
from ._sf import ShapeletForestCounterfactual

__all__ = [
    "counterfactuals",
    "proximity",
    "ShapeletForestCounterfactual",
    "KNeighborsCounterfactual",
    "PrototypeCounterfactual",
]



# Authors: Isak Samsten
# License: BSD 3 clause


from sklearn.utils._param_validation import StrOptions

from ..transform._interval import IntervalMixin
from ..transform._pivot import PivotMixin
from ..transform._rocket import RocketMixin
from ..transform._shapelet import DilatedShapeletMixin, ShapeletMixin
from ..tree._ctree import (
    DynamicTreeAttributeGenerator,
    EntropyCriterion,
    ExtraTreeBuilder,
    GiniCriterion,
    MSECriterion,
    Tree,
    TreeAttributeGenerator,
    TreeBuilder,
)
from ..utils.validation import check_option
from ._base import BaseTree, BaseTreeClassifier, BaseTreeRegressor

CLF_CRITERION = {"gini": GiniCriterion, "entropy": EntropyCriterion}
REG_CRITERION = {"squared_error": MSECriterion}


class FeatureTreeMixin:
    def _wrap_generator(self, generator):
        return TreeAttributeGenerator(generator)

    def _fit(self, x, y, sample_weights, max_depth, random_state):
        generator = self._wrap_generator(self._get_generator(x, y))
        tree_builder = self._get_tree_builder(
            x,
            y,
            sample_weights,
            generator,
            random_state,
            max_depth,
        )
        tree_builder.build_tree()
        self.tree_ = tree_builder.tree_


class BaseFeatureTreeRegressor(FeatureTreeMixin, BaseTreeRegressor):
    _parameter_constraints: dict = {
        **BaseTree._parameter_constraints,
        "criterion": [
            StrOptions(REG_CRITERION.keys()),
        ],
    }

    def _get_tree_builder(
        self, x, y, sample_weights, generator, random_state, max_depth
    ):
        Criterion = REG_CRITERION[self.criterion]
        return TreeBuilder(
            x,
            sample_weights,
            generator,
            Criterion(y),
            Tree(generator, 1),
            random_state,
            max_depth=max_depth,
            min_sample_split=self.min_samples_split,
            min_sample_leaf=self.min_samples_leaf,
            min_impurity_decrease=self.min_impurity_decrease,
        )


class BaseFeatureTreeClassifier(FeatureTreeMixin, BaseTreeClassifier):
    _parameter_constraints: dict = {
        **BaseTree._parameter_constraints,
        "criterion": [
            StrOptions(CLF_CRITERION.keys()),
        ],
        "class_weight": [
            StrOptions({"balanced"}),
            dict,
            None,
        ],
    }

    def _get_tree_builder(
        self, x, y, sample_weights, generator, random_state, max_depth
    ):
        Criterion = CLF_CRITERION[self.criterion]
        return TreeBuilder(
            x,
            sample_weights,
            generator,
            Criterion(y, self.n_classes_),
            Tree(generator, self.n_classes_),
            random_state,
            max_depth=max_depth,
            min_sample_split=self.min_samples_split,
            min_sample_leaf=self.min_samples_leaf,
            min_impurity_decrease=self.min_impurity_decrease,
        )


class DynamicTreeMixin:
    _parameter_constraints: dict = {"alpha": [float, None]}

    def _wrap_generator(self, generator):
        if hasattr(self, "alpha") and self.alpha is not None:
            if self.alpha == 0.0:
                raise ValueError("alpha == 0.0, must be != 0")

            return DynamicTreeAttributeGenerator(generator, self.alpha)

        return TreeAttributeGenerator(generator)


class ShapeletTreeRegressor(DynamicTreeMixin, ShapeletMixin, BaseFeatureTreeRegressor):
    """
    A shapelet tree regressor.

    Parameters
    ----------
    max_depth : int, optional
        The maximum depth of the tree. If ``None`` the tree is
        expanded until all leaves are pure or until all leaves contain less
        than ``min_samples_split`` samples.
    min_samples_split : int, optional
        The minimum number of samples to split an internal node.
    min_samples_leaf : int, optional
        The minimum number of samples in a leaf.
    min_impurity_decrease : float, optional
        A split will be introduced only if the impurity decrease is larger
        than or equal to this value.
    n_shapelets : int, optional
        The number of shapelets to sample at each node.
    min_shapelet_size : float, optional
        The minimum length of a shapelets expressed as a fraction of
        *n_timestep*.
    max_shapelet_size : float, optional
        The maximum length of a shapelets expressed as a fraction of
        *n_timestep*.
    alpha : float, optional
        Dynamically decrease the number of sampled shapelets at each node according
        to the current depth, i.e.:

        ::
            w = 1 - exp(-abs(alpha) * depth)

        - if ``alpha < 0``, the number of sampled shapelets decrease from
            ``n_shapelets`` towards 1 with increased depth.
        - if ``alpha > 0``, the number of sampled shapelets increase from ``1``
            towards ``n_shapelets`` with increased depth.
        - if ``None``, the number of sampled shapelets are the same
            independeth of depth.
    metric : str or list, optional
        - If ``str``, the distance metric used to identify the best
            shapelet.
        - If ``list``, multiple metrics specified as a list of
            tuples, where the first element of the tuple is a metric name and
            the second element a dictionary with a parameter grid
            specification. A parameter grid specification is a dict with two
            mandatory and one optional key-value pairs defining the lower and
            upper bound on the values and number of values in the grid. For
            example, to specifiy a grid over the argument ``r`` with 10
            values in the range 0 to 1, we would give the following
            specification: ``dict(min_r=0, max_r=1, num_r=10)``.

            Read more about metric specifications in the `User guide
            <metric_specification>`__

        .. versionchanged:: 1.2
            Added support for multi-metric shapelet transform
    metric_params : dict, optional
        Parameters for the distance measure. Ignored unless metric is a string.

        Read more about the parameters in the `User guide
        <list_of_subsequence_metrics>`__.
    criterion : {"squared_error"}, optional
        The criterion used to evaluate the utility of a split.

        .. deprecated:: 1.1
            Criterion "mse" was deprecated in v1.1 and removed in version 1.2.
    random_state : int or RandomState
        - If ``int``, ``random_state`` is the seed used by the
            random number generator
        - If :class:`numpy.random.RandomState` instance, ``random_state``
            is the random number generator
        - If ``None``, the random number generator is the
            :class:`numpy.random.RandomState` instance used by
            :func:`numpy.random`.

    Attributes
    ----------
    tree_ : Tree
        The internal tree representation
    """

    _parameter_constraints: dict = {
        **BaseFeatureTreeRegressor._parameter_constraints,
        **ShapeletMixin._parameter_constraints,
        **DynamicTreeMixin._parameter_constraints,
        "random_state": ["random_state"],
    }

    def __init__(
        self,
        *,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        n_shapelets="log2",
        min_shapelet_size=0,
        max_shapelet_size=1,
        alpha=None,
        metric="euclidean",
        metric_params=None,
        criterion="squared_error",
        random_state=None,
    ):
        super().__init__(
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
        )
        self.random_state = random_state
        self.n_shapelets = n_shapelets
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.metric = metric
        self.metric_params = metric_params
        self.criterion = criterion
        self.alpha = alpha


class ExtraShapeletTreeRegressor(ShapeletTreeRegressor):
    """
    An extra shapelet tree regressor.

    Extra shapelet trees are constructed by sampling a distance threshold
    uniformly in the range [min(dist), max(dist)].

    Parameters
    ----------
    n_shapelets : int, optional
        The number of shapelets to sample at each node.
    max_depth : int, optional
        The maximum depth of the tree. If `None` the tree is expanded until all
        leaves are pure or until all leaves contain less than `min_samples_split`
        samples.
    min_samples_split : int, optional
        The minimum number of samples to split an internal node.
    min_samples_leaf : int, optional
        The minimum number of samples in a leaf.
    criterion : {"squared_error"}, optional
        The criterion used to evaluate the utility of a split.

        .. deprecated:: 1.1
            Criterion "mse" was deprecated in v1.1 and removed in version 1.2.
    min_impurity_decrease : float, optional
        A split will be introduced only if the impurity decrease is larger than or
        equal to this value.
    min_shapelet_size : float, optional
        The minimum length of a sampled shapelet expressed as a fraction, computed
        as `min(ceil(X.shape[-1] * min_shapelet_size), 2)`.
    max_shapelet_size : float, optional
        The maximum length of a sampled shapelet, expressed as a fraction, computed
        as `ceil(X.shape[-1] * max_shapelet_size)`.
    metric : {'euclidean', 'scaled_euclidean', 'scaled_dtw'}, optional
        Distance metric used to identify the best shapelet.
    metric_params : dict, optional
        Parameters for the distance measure.
    random_state : int or RandomState
        - If `int`, `random_state` is the seed used by the random number generator;
        - If `RandomState` instance, `random_state` is the random number generator;
        - If `None`, the random number generator is the `RandomState` instance used
            by `np.random`.

    Attributes
    ----------
    tree_ : Tree
        The internal tree representation

    """

    _parameter_constraints: dict = {**ShapeletTreeRegressor._parameter_constraints}
    _parameter_constraints.pop("alpha")

    def __init__(
        self,
        *,
        n_shapelets=1,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        min_shapelet_size=0.0,
        max_shapelet_size=1.0,
        metric="euclidean",
        metric_params=None,
        criterion="squared_error",
        random_state=None,
    ):
        super(ExtraShapeletTreeRegressor, self).__init__(
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
            n_shapelets=n_shapelets,
            min_shapelet_size=min_shapelet_size,
            max_shapelet_size=max_shapelet_size,
            metric=metric,
            metric_params=metric_params,
            criterion=criterion,
            random_state=random_state,
        )

    def _get_tree_builder(
        self, x, y, sample_weights, generator, random_state, max_depth
    ):
        Criterion = check_option(REG_CRITERION, self.criterion, "criterion")
        return ExtraTreeBuilder(
            x,
            sample_weights,
            generator,
            Criterion(y),
            Tree(generator, 1),
            random_state,
            max_depth=max_depth,
            min_sample_split=self.min_samples_split,
            min_sample_leaf=self.min_samples_leaf,
            min_impurity_decrease=self.min_impurity_decrease,
        )


class ShapeletTreeClassifier(
    DynamicTreeMixin, ShapeletMixin, BaseFeatureTreeClassifier
):
    """
    A shapelet tree classifier.

    Parameters
    ----------
    n_shapelets : int, optional
        The number of shapelets to sample at each node.
    max_depth : int, optional
        The maximum depth of the tree. If `None` the tree is expanded until all
        leaves are pure or until all leaves contain less than `min_samples_split`
        samples.
    min_samples_split : int, optional
        The minimum number of samples to split an internal node.
    min_samples_leaf : int, optional
        The minimum number of samples in a leaf.
    criterion : {"entropy", "gini"}, optional
        The criterion used to evaluate the utility of a split.
    min_impurity_decrease : float, optional
        A split will be introduced only if the impurity decrease is larger than or
        equal to this value.
    min_shapelet_size : float, optional
        The minimum length of a sampled shapelet expressed as a fraction, computed
        as ``min(ceil(X.shape[-1] * min_shapelet_size), 2)``.
    max_shapelet_size : float, optional
        The maximum length of a sampled shapelet, expressed as a fraction, computed
        as ``ceil(X.shape[-1] * max_shapelet_size)``.
    alpha : float, optional
        Dynamically decrease the number of sampled shapelets at each node according
        to the current depth.

        .. math:`w = 1 - e^{-|alpha| * depth})`

        - if :math:`alpha < 0`, the number of sampled shapelets decrease from
            ``n_shapelets`` towards 1 with increased depth.

            .. math:`n_shapelets * (1 - w)`
        - if :math:`alpha > 0`, the number of sampled shapelets increase from ``1``
            towards ``n_shapelets`` with increased depth.

            .. math:`n_shapelets * w`
        - if ``None``, the number of sampled shapelets are the same independeth of
            depth.

    metric : {"euclidean", "scaled_euclidean", "dtw", "scaled_dtw"}, optional
        Distance metric used to identify the best shapelet.
    metric_params : dict, optional
        Parameters for the distance measure
    class_weight : dict or "balanced", optional
        Weights associated with the labels

        - if dict, weights on the form {label: weight}
        - if "balanced" each class weight inversely proportional to the class
            frequency
        - if None, each class has equal weight.
    random_state : int or RandomState
        - If `int`, `random_state` is the seed used by the random number generator;
        - If `RandomState` instance, `random_state` is the random number generator;
        - If `None`, the random number generator is the `RandomState` instance used
            by `np.random`.

    Attributes
    ----------
    tree_ : Tree
        The tree data structure used internally
    classes_ : ndarray of shape (n_classes,)
        The class labels
    n_classes_ : int
        The number of class labels

    See Also
    --------
    ShapeletTreeRegressor : A shapelet tree regressor.
    ExtraShapeletTreeClassifier : An extra random shapelet tree classifier.
    """

    _parameter_constraints: dict = {
        **BaseFeatureTreeClassifier._parameter_constraints,
        **ShapeletMixin._parameter_constraints,
        **DynamicTreeMixin._parameter_constraints,
        "random_state": ["random_state"],
    }

    def __init__(
        self,
        *,
        n_shapelets="log2",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        min_shapelet_size=0.0,
        max_shapelet_size=1.0,
        alpha=None,
        metric="euclidean",
        metric_params=None,
        criterion="entropy",
        class_weight=None,
        random_state=None,
    ):
        super().__init__(
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
        )
        self.random_state = random_state
        self.n_shapelets = n_shapelets
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.metric = metric
        self.metric_params = metric_params
        self.criterion = criterion
        self.class_weight = class_weight
        self.alpha = alpha


class ExtraShapeletTreeClassifier(ShapeletTreeClassifier):
    """
    An extra shapelet tree classifier.

    Extra shapelet trees are constructed by sampling a distance threshold
    uniformly in the range ``[min(dist), max(dist)]``.

    Parameters
    ----------
    n_shapelets : int, optional
        The number of shapelets to sample at each node.
    max_depth : int, optional
        The maximum depth of the tree. If `None` the tree is expanded until all
        leaves are pure or until all leaves contain less than `min_samples_split`
        samples.
    min_samples_leaf : int, optional
        The minimum number of samples in a leaf.
    min_impurity_decrease : float, optional
        A split will be introduced only if the impurity decrease is larger than or
        equal to this value.
    min_samples_split : int, optional
        The minimum number of samples to split an internal node.
    min_shapelet_size : float, optional
        The minimum length of a sampled shapelet expressed as a fraction, computed
        as `min(ceil(X.shape[-1] * min_shapelet_size), 2)`.
    max_shapelet_size : float, optional
        The maximum length of a sampled shapelet, expressed as a fraction, computed
        as `ceil(X.shape[-1] * max_shapelet_size)`.
    metric : {"euclidean", "scaled_euclidean", "dtw", "scaled_dtw"}, optional
        Distance metric used to identify the best shapelet.
    metric_params : dict, optional
        Parameters for the distance measure.
    criterion : {"entropy", "gini"}, optional
        The criterion used to evaluate the utility of a split.
    class_weight : dict or "balanced", optional
        Weights associated with the labels

        - if dict, weights on the form {label: weight}
        - if "balanced" each class weight inversely proportional to the class
            frequency
        - if None, each class has equal weight.
    random_state : int or RandomState, optional
        - If `int`, `random_state` is the seed used by the random number generator;
        - If `RandomState` instance, `random_state` is the random number generator;
        - If `None`, the random number generator is the `RandomState` instance used
            by `np.random`.

    Attributes
    ----------
    tree_ : Tree
        The tree representation
    """

    _parameter_constraints: dict = {**ShapeletTreeClassifier._parameter_constraints}
    _parameter_constraints.pop("alpha")

    def __init__(
        self,
        *,
        n_shapelets=1,
        max_depth=None,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        min_samples_split=2,
        min_shapelet_size=0.0,
        max_shapelet_size=1.0,
        metric="euclidean",
        metric_params=None,
        criterion="entropy",
        class_weight=None,
        random_state=None,
    ):
        super(ExtraShapeletTreeClassifier, self).__init__(
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
            n_shapelets=n_shapelets,
            min_shapelet_size=min_shapelet_size,
            max_shapelet_size=max_shapelet_size,
            metric=metric,
            metric_params=metric_params,
            criterion=criterion,
            class_weight=class_weight,
            random_state=random_state,
        )

    def _get_tree_builder(
        self, x, y, sample_weights, generator, random_state, max_depth
    ):
        Criterion = check_option(CLF_CRITERION, self.criterion, "criterion")
        return ExtraTreeBuilder(
            x,
            sample_weights,
            generator,
            Criterion(y, self.n_classes_),
            Tree(generator, self.n_classes_),
            random_state,
            max_depth=max_depth,
            min_sample_split=self.min_samples_split,
            min_sample_leaf=self.min_samples_leaf,
            min_impurity_decrease=self.min_impurity_decrease,
        )


class RocketTreeRegressor(RocketMixin, BaseFeatureTreeRegressor):
    """
    A tree regressor that uses random convolutions as features.

    Attributes
    ----------
    tree_ : Tree
        The internal tree representation.
    """

    _parameter_constraints: dict = {
        **BaseFeatureTreeRegressor._parameter_constraints,
        **RocketMixin._parameter_constraints,
        "random_state": ["random_state"],
    }

    def __init__(
        self,
        n_kernels=10,
        *,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        criterion="squared_error",
        sampling="normal",
        sampling_params=None,
        kernel_size=None,
        bias_prob=1.0,
        normalize_prob=1.0,
        padding_prob=0.5,
        random_state=None,
    ):
        """Construct a new rocket tree regressor.

        Parameters
        ----------
        n_kernels : int, optional
            The number of kernels to sample at each node.
        max_depth : int, optional
            The maximum depth of the tree. If `None` the tree is expanded until all
            leaves are pure or until all leaves contain less than `min_samples_split`
            samples.
        min_samples_split : int, optional
            The minimum number of samples to split an internal node.
        min_samples_leaf : int, optional
            The minimum number of samples in a leaf.
        min_impurity_decrease : float, optional
            A split will be introduced only if the impurity decrease is larger than or
            equal to this value.
        criterion : {"entropy", "gini"}, optional
            The criterion used to evaluate the utility of a split.
        kernel_size : array-like, optional
            The kernel size, by default ``[7, 11, 13]``.
        min_size : float, optional
            The minimum timestep fraction to generate kernel sizes. If set,
            ``kernel_size`` cannot be set.
        max_size : float, optional
            The maximum timestep fractio to generate kernel sizes, If set,
            ``kernel_size`` cannot be set.
        bias_prob : float, optional
            The probability of using a bias term.
        normalize_prob : float, optional
            The probability of performing normalization.
        padding_prob : float, optional
            The probability of padding with zeros.
        random_state : int or RandomState
            - If `int`, `random_state` is the seed used by the random number generator
            - If `RandomState` instance, `random_state` is the random number generator
            - If `None`, the random number generator is the `RandomState` instance used
              by `np.random`.
        """
        super().__init__(
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
        )
        self.n_kernels = n_kernels
        self.criterion = criterion
        self.sampling = sampling
        self.sampling_params = sampling_params
        self.kernel_size = kernel_size
        self.bias_prob = bias_prob
        self.normalize_prob = normalize_prob
        self.padding_prob = padding_prob
        self.random_state = random_state


class RocketTreeClassifier(RocketMixin, BaseFeatureTreeClassifier):
    """A tree classifier that uses random convolutions as features.

    Attributes
    ----------
    tree_ : Tree
        The internal tree representation.

    """

    _parameter_constraints: dict = {
        **BaseFeatureTreeClassifier._parameter_constraints,
        **RocketMixin._parameter_constraints,
        "random_state": ["random_state"],
    }

    def __init__(
        self,
        n_kernels=10,
        *,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        criterion="entropy",
        sampling="normal",
        sampling_params=None,
        kernel_size=None,
        min_size=None,
        max_size=None,
        bias_prob=1.0,
        normalize_prob=1.0,
        padding_prob=0.5,
        class_weight=None,
        random_state=None,
    ):
        """Construct a new rocket tree classifier.

        Parameters
        ----------
        n_kernels : int, optional
            The number of kernels to sample at each node.
        max_depth : int, optional
            The maximum depth of the tree. If `None` the tree is expanded until all
            leaves are pure or until all leaves contain less than `min_samples_split`
            samples.
        min_samples_split : int, optional
            The minimum number of samples to split an internal node.
        min_samples_leaf : int, optional
            The minimum number of samples in a leaf.
        min_impurity_decrease : float, optional
            A split will be introduced only if the impurity decrease is larger than or
            equal to this value.
        criterion : {"entropy", "gini"}, optional
            The criterion used to evaluate the utility of a split.
        sampling : {"normal", "uniform", "shapelet"}, optional
            The sampling of convolutional filters.

            - if "normal", sample filter according to a normal distribution with
              ``mean`` and ``scale``.
            - if "uniform", sample filter according to a uniform distribution with
              ``lower`` and ``upper``.
            - if "shapelet", sample filters as subsequences in the training data.
        sampling_params : dict, optional
            The parameters for the sampling.

            - if "normal", ``{"mean": float, "scale": float}``, defaults to
               ``{"mean": 0, "scale": 1}``.
            - if "uniform", ``{"lower": float, "upper": float}``, defaults to
               ``{"lower": -1, "upper": 1}``.
        kernel_size : array-like, optional
            The kernel size, by default ``[7, 11, 13]``.
        min_size : float, optional
            The minimum timestep fraction to generate kernel sizes. If set,
            ``kernel_size`` cannot be set.
        max_size : float, optional
            The maximum timestep fractio to generate kernel sizes, If set,
            ``kernel_size`` cannot be set.
        bias_prob : float, optional
            The probability of using a bias term.
        normalize_prob : float, optional
            The probability of performing normalization.
        padding_prob : float, optional
            The probability of padding with zeros.
        class_weight : dict or "balanced", optional
            Weights associated with the labels

            - if dict, weights on the form {label: weight}
            - if "balanced" each class weight inversely proportional to the class
              frequency
            - if None, each class has equal weight
        random_state : int or RandomState
            - If `int`, `random_state` is the seed used by the random number generator
            - If `RandomState` instance, `random_state` is the random number generator
            - If `None`, the random number generator is the `RandomState` instance used
              by `np.random`.
        """
        super().__init__(
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
        )
        self.n_kernels = n_kernels
        self.criterion = criterion
        self.sampling = sampling
        self.sampling_params = sampling_params
        self.kernel_size = kernel_size
        self.min_size = min_size
        self.max_size = max_size
        self.bias_prob = bias_prob
        self.normalize_prob = normalize_prob
        self.padding_prob = padding_prob
        self.random_state = random_state
        self.class_weight = class_weight


class IntervalTreeClassifier(IntervalMixin, BaseFeatureTreeClassifier):
    """An interval based tree classifier.

    Attributes
    ----------
    tree_ : Tree
        The internal tree structure.

    """

    _parameter_constraints: dict = {
        **BaseFeatureTreeClassifier._parameter_constraints,
        **IntervalMixin._parameter_constraints,
        "random_state": ["random_state"],
    }

    def __init__(
        self,
        n_intervals="sqrt",
        *,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        criterion="entropy",
        intervals="fixed",
        sample_size=0.5,
        min_size=0.0,
        max_size=1.0,
        summarizer="mean_var_slope",
        class_weight=None,
        random_state=None,
    ):
        """Construct a new interval tree classifier.

        Parameters
        ----------
        n_intervals : {"log", "sqrt"}, int or float, optional
            The number of intervals to partition the time series into.

            - if "log", the number of intervals is ``log2(n_timestep)``.
            - if "sqrt", the number of intervals is ``sqrt(n_timestep)``.
            - if int, the number of intervals is ``n_intervals``.
            - if float, the number of intervals is ``n_intervals * n_timestep``, with
              ``0 < n_intervals < 1``.
        max_depth : int, optional
            The maximum depth of the tree. If `None` the tree is expanded until all
            leaves are pure or until all leaves contain less than `min_samples_split`
            samples.
        min_samples_split : int, optional
            The minimum number of samples to split an internal node.
        min_samples_leaf : int, optional
            The minimum number of samples in a leaf.
        min_impurity_decrease : float, optional
            A split will be introduced only if the impurity decrease is larger than or
            equal to this value.
        criterion : {"entropy", "gini"}, optional
            The criterion used to evaluate the utility of a split.
        intervals : {"fixed", "sample", "random"}, optional

            - if "fixed", `n_intervals` non-overlapping intervals.
            - if "sample", ``n_intervals * sample_size`` non-overlapping intervals.
            - if "random", `n_intervals` possibly overlapping intervals of randomly
              sampled in ``[min_size * n_timestep, max_size * n_timestep]``
        sample_size : float, optional
            The fraction of intervals to sample at each node. Ignored unless
            ``intervals="sample"``.
        min_size : float, optional
            The minmum interval size. Ignored unless ``intervals="random"``.
        max_size : float, optional
            The maximum interval size. Ignored unless ``intervals="random"``.
        summarizer : list or str, optional
            The summarization of each interval.

            - if list, a list of callables accepting a numpy array returing a float.
            - if str, a predified summarized. See
              :mod:`wildboar.transform._interval._INTERVALS.keys()` for all supported
              summarizers.
        class_weight : dict or "balanced", optional
            Weights associated with the labels
            - if dict, weights on the form {label: weight}
            - if "balanced" each class weight inversely proportional to the class
              frequency
            - if None, each class has equal weight
        random_state : int or RandomState
            - If `int`, `random_state` is the seed used by the random number generator
            - If `RandomState` instance, `random_state` is the random number generator
            - If `None`, the random number generator is the `RandomState` instance used
              by `np.random`.
        """
        super().__init__(
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
        )
        self.n_intervals = n_intervals
        self.intervals = intervals
        self.sample_size = sample_size
        self.min_size = min_size
        self.max_size = max_size
        self.summarizer = summarizer
        self.random_state = random_state
        self.class_weight = class_weight
        self.criterion = criterion


class IntervalTreeRegressor(IntervalMixin, BaseFeatureTreeRegressor):
    """An interval based tree regressor.

    Attributes
    ----------
    tree_ : Tree
        The internal tree structure.

    """

    _parameter_constraints: dict = {
        **BaseFeatureTreeRegressor._parameter_constraints,
        **IntervalMixin._parameter_constraints,
        "random_state": ["random_state"],
    }

    def __init__(
        self,
        n_intervals="sqrt",
        *,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        criterion="squared_error",
        intervals="fixed",
        sample_size=0.5,
        min_size=0.0,
        max_size=1.0,
        summarizer="mean_var_slope",
        random_state=None,
    ):
        """Construct a new interval tree regressor.

        Parameters
        ----------
        n_intervals : {"log", "sqrt"}, int or float, optional
            The number of intervals to partition the time series into.

            - if "log", the number of intervals is ``log2(n_timestep)``.
            - if "sqrt", the number of intervals is ``sqrt(n_timestep)``.
            - if int, the number of intervals is ``n_intervals``.
            - if float, the number of intervals is ``n_intervals * n_timestep``, with
              ``0 < n_intervals < 1``.
        max_depth : int, optional
            The maximum depth of the tree. If `None` the tree is expanded until all
            leaves are pure or until all leaves contain less than `min_samples_split`
            samples.
        min_samples_split : int, optional
            The minimum number of samples to split an internal node.
        min_samples_leaf : int, optional
            The minimum number of samples in a leaf.
        min_impurity_decrease : float, optional
            A split will be introduced only if the impurity decrease is larger than or
            equal to this value.
        criterion : {"squared_error"}, optional
            The criterion used to evaluate the utility of a split.

            .. deprecated:: 1.1
                Criterion "mse" was deprecated in v1.1 and removed in version 1.2.
        intervals : {"fixed", "sample", "random"}, optional
            - if "fixed", `n_intervals` non-overlapping intervals.
            - if "sample", ``n_intervals * sample_size`` non-overlapping intervals.
            - if "random", `n_intervals` possibly overlapping intervals of randomly
              sampled in ``[min_size * n_timestep, max_size * n_timestep]``
        sample_size : float, optional
            The fraction of intervals to sample at each node. Ignored unless
            ``intervals="sample"``.
        min_size : float, optional
            The minmum interval size. Ignored unless ``intervals="random"``.
        max_size : float, optional
            The maximum interval size. Ignored unless ``intervals="random"``.
        summarizer : list or str, optional
            The summarization of each interval.

            - if list, a list of callables accepting a numpy array returing a float.
            - if str, a predified summarized. See
              :mod:`wildboar.transform._interval._INTERVALS.keys()` for all supported
              summarizers.
        random_state : int or RandomState
            - If `int`, `random_state` is the seed used by the random number generator
            - If `RandomState` instance, `random_state` is the random number generator
            - If `None`, the random number generator is the `RandomState` instance used
              by `np.random`.
        """
        super().__init__(
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
        )
        self.n_intervals = n_intervals
        self.intervals = intervals
        self.sample_size = sample_size
        self.min_size = min_size
        self.max_size = max_size
        self.summarizer = summarizer
        self.random_state = random_state
        self.criterion = criterion


class PivotTreeClassifier(PivotMixin, BaseFeatureTreeClassifier):
    """A tree classifier that uses pivot time series.

    Attributes
    ----------
    tree_ : Tree
        The internal tree representation

    """

    _parameter_constraints: dict = {
        **BaseFeatureTreeClassifier._parameter_constraints,
        **PivotMixin._parameter_constraints,
        "random_state": ["random_state"],
    }

    def __init__(
        self,
        n_pivot="sqrt",
        *,
        metrics="all",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        criterion="entropy",
        class_weight=None,
        random_state=None,
    ):
        """Construct a new pivot tree classifier.

        Parameters
        ----------
        n_pivot : str or int, optional
            The number of pivot time series to sample at each node.
        metrics : str, optional
            The metrics to sample from. Currently, we only support "all".
        max_depth : int, optional
            The maximum depth of the tree. If `None` the tree is expanded until all
            leaves are pure or until all leaves contain less than `min_samples_split`
            samples.
        min_samples_split : int, optional
            The minimum number of samples to split an internal node.
        min_samples_leaf : int, optional
            The minimum number of samples in a leaf.
        min_impurity_decrease : float, optional
            A split will be introduced only if the impurity decrease is larger than or
            equal to this value.
        criterion : {"entropy", "gini"}, optional
            The criterion used to evaluate the utility of a split.
        class_weight : dict or "balanced", optional
            Weights associated with the labels.

            - if dict, weights on the form {label: weight}.
            - if "balanced" each class weight inversely proportional to the class
              frequency.
            - if None, each class has equal weight.
        random_state : int or RandomState
            - If `int`, `random_state` is the seed used by the random number generator
            - If `RandomState` instance, `random_state` is the random number generator
            - If `None`, the random number generator is the `RandomState` instance used
              by `np.random`.
        """
        super().__init__(
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
        )
        self.n_pivot = n_pivot
        self.metrics = metrics
        self.random_state = random_state
        self.criterion = criterion
        self.class_weight = class_weight


class DilatedShapeletTreeClassifier(DilatedShapeletMixin, BaseFeatureTreeClassifier):
    _parameter_constraints: dict = {
        **BaseFeatureTreeClassifier._parameter_constraints,
        **DilatedShapeletMixin._parameter_constraints,
        "random_state": ["random_state"],
    }

    def __init__(
        self,
        n_shapelets=1000,
        *,
        metric="euclidean",
        metric_params=None,
        shapelet_size=None,
        min_shapelet_size=None,
        max_shapelet_size=None,
        normalize_prob=0.8,
        lower=0.05,
        upper=0.1,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        criterion="entropy",
        class_weight=None,
        random_state=None,
    ):
        super().__init__(
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
        )
        self.n_shapelets = n_shapelets
        self.metric = metric
        self.metric_params = metric_params
        self.shapelet_size = shapelet_size
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.normalize_prob = normalize_prob
        self.lower = lower
        self.upper = upper
        self.criterion = criterion
        self.class_weight = class_weight
        self.random_state = random_state



# Authors: Isak Samsten
# License: BSD 3 clause

from abc import ABCMeta, abstractmethod

import numpy as np
from sklearn.base import ClassifierMixin, RegressorMixin
from sklearn.linear_model import RidgeClassifierCV, RidgeCV
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.utils.extmath import softmax
from sklearn.utils.validation import check_is_fitted, check_random_state

from ..base import BaseEstimator


class BaseTransformEstimator(BaseEstimator, metaclass=ABCMeta):
    def __init__(self, *, random_state=None, n_jobs=None):
        self.random_state = random_state
        self.n_jobs = n_jobs

    def fit(self, x, y, sample_weight=None):
        x, y = self._validate_data(x, y, dtype=float, allow_3d=True)
        self.pipe_ = Pipeline(self._build_pipeline())
        self.pipe_.fit(x, y, estimator__sample_weight=sample_weight)
        return self

    def _build_pipeline(self):
        random_state = check_random_state(self.random_state)
        return [
            ("transform", self._get_transform(random_state.randint(2**31))),
            ("estimator", self._get_estimator(random_state.randint(2**31))),
        ]

    @abstractmethod
    def _get_transform(self, random_state):
        pass

    @abstractmethod
    def _get_estimator(self, random_state):
        pass


class BaseTransformClassifier(ClassifierMixin, BaseTransformEstimator):
    def predict(self, x):
        check_is_fitted(self)
        x = self._validate_data(x, dtype=float, allow_3d=True)
        return self.pipe_.predict(x)

    def predict_proba(self, x):
        check_is_fitted(self)
        x = self._validate_data(x, dtype=float, allow_3d=True)
        return self.pipe_.predict_proba(x)

    def predict_log_proba(self, x):
        check_is_fitted(self)
        x = self._validate_data(x, dtype=float, allow_3d=True)
        return self.pipe_.predict_log_proba(x)

    def decision_function(self, x):
        check_is_fitted(self)
        x = self._validate_data(x, dtype=float, allow_3d=True)
        return self.pipe_.decision_function(x)

    @property
    def classes_(self):
        return self.pipe_.classes_


class BaseTransformRegressor(RegressorMixin, BaseTransformEstimator):
    def predict(self, x):
        check_is_fitted(self)
        x = self._validate_data(x, dtype=float, allow_3d=True)
        return self.pipe_.predict(x)

    def decision_function(self, x):
        check_is_fitted(self)
        x = self._validate_data(x, dtype=float, allow_3d=True)
        return self.pipe_.decision_function(x)


class TransformRidgeClassifierCV(BaseTransformClassifier):
    def __init__(
        self,
        *,
        alphas=(0.1, 1.0, 10.0),
        fit_intercept=True,
        scoring=None,
        cv=None,
        class_weight=None,
        normalize=True,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(random_state=random_state, n_jobs=n_jobs)
        self.alphas = alphas
        self.fit_intercept = fit_intercept
        self.scoring = scoring
        self.cv = cv
        self.normalize = normalize
        self.class_weight = class_weight
        self.random_state = random_state

    def _get_estimator(self, random_state):
        return RidgeClassifierCV(
            alphas=self.alphas,
            fit_intercept=self.fit_intercept,
            scoring=self.scoring,
            cv=self.cv,
            class_weight=self.class_weight,
            store_cv_values=False,
        )

    def _build_pipeline(self):
        pipeline = super()._build_pipeline()
        if self.normalize is not False:
            pipeline.insert(1, ("normalize", StandardScaler()))

        return pipeline

    def predict_proba(self, x):
        decision = self.decision_function(x)
        if decision.ndim == 1:
            decision_2d = np.c_[-decision, decision]
        else:
            decision_2d = decision
        return softmax(decision_2d, copy=False)


class TransformRidgeCV(BaseTransformRegressor):
    def __init__(
        self,
        *,
        alphas=(0.1, 1.0, 10.0),
        fit_intercept=True,
        scoring=None,
        cv=None,
        gcv_mode=None,
        normalize=True,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(random_state=random_state, n_jobs=n_jobs)
        self.alphas = np.asarray(alphas)
        self.fit_intercept = fit_intercept
        self.scoring = scoring
        self.cv = cv
        self.normalize = normalize
        self.gcv_mode = gcv_mode

    def _get_estimator(self, random_state):
        return RidgeCV(
            alphas=self.alphas,
            fit_intercept=self.fit_intercept,
            scoring=self.scoring,
            cv=self.cv,
            gcv_mode=self.gcv_mode,
            store_cv_values=False,
            alpha_per_target=False,
        )

    def _build_pipeline(self):
        pipeline = super()._build_pipeline()
        if self.normalize:
            pipeline.insert(1, StandardScaler())

        return pipeline



# Authors: Isak Samsten
# License: BSD 3 clause

"""
Annotation of time series.

See the :ref:`annotation section in the User Guide
<guide-annotate>` for more details and examples.
"""
from ._motifs import motifs
from ._segment import segment

__all__ = ["motifs", "segment"]



# Authors: Isak Samsten
# License: BSD 3 clause

"""
Dataset loading utilities.

See the dataset section in the :ref:`User Guide <guide-datasets>` for
more details and examples.

Examples
--------
>>> from wildboar.datasets import load_dataset
>>> X, y = load_dataset("GunPoint")
>>> X.shape
(200, 60)
"""

import re

import numpy as np

from wildboar.utils.validation import check_option

from ..utils import os_cache_path
from ._filter import make_filter
from ._repository import (
    Bundle,
    JSONRepository,
    NpBundle,
    Repository,
    RepositoryCollection,
)
from .preprocess import _PREPROCESS

__all__ = [
    "Repository",
    "JSONRepository",
    "Bundle",
    "NpBundle",
    "set_cache_dir",
    "get_repository",
    "clear_cache",
    "list_repositories",
    "refresh_repositories",
    "list_bundles",
    "get_bundles",
    "install_repository",
    "load_dataset",
    "list_datasets",
    "load_datasets",
    "list_collections",
    "load_two_lead_ecg",
    "load_synthetic_control",
    "load_gun_point",
]


def _split_repo_bundle(repo_bundle_name):
    """
    Split a repository bundle string of the format {repo}/{bundle}[:tag].

    Parameters
    ----------
    repo_bundle_name : str
        {repo}/{bundle} name.

    Returns
    -------
    repository : str
        Key of the repository.
    bundle : str
        Name of the bundle.
    tag : str
        An optional tag.
    """
    match = re.match(
        r"(?:([a-zA-Z]+)/([a-zA-Z0-9\-]+))?(?::([a-zA-Z\-]+))?$",  # noqa: E501
        repo_bundle_name,
    )
    if match:
        repository_name = match.group(1)
        bundle_name = match.group(2)
        tag = match.group(3)
        return repository_name, bundle_name, tag
    else:
        raise ValueError("Invalid repository/bundle string '%s'." % repo_bundle_name)


# noqa: H0002
def set_cache_dir(cache_dir=None):
    """
    Change the global cache directory.

    If called without arguments, the cache directory is reset to the default
    directory.

    Parameters
    ----------
    cache_dir : str, optional
        The cache directory root.
    """
    global _CACHE_DIR  # noqa: PLW0603
    _CACHE_DIR = cache_dir


def _default_cache_dir():
    return os_cache_path("wildboar") if not _CACHE_DIR else _CACHE_DIR


# noqa: H0002
def load_synthetic_control(merge_train_test=True):
    """
    Load the Synthetic_Control dataset.

    Parameters
    ----------
    merge_train_test : bool, optional
        Merge the existing training and testing partitions.

    Returns
    -------
    x : ndarray, optional
        The samples if `merge_train_test=False`.
    y : ndarray, optional
        The labels, if `merge_train_test=False`.
    x_train : ndarray, optional
        The training samples if `merge_train_test=False`.
    x_test : ndarray, optional
        The testing samples if `merge_train_test=False`.
    y_train : ndarray, optional
        The training labels if `merge_train_test=False`.
    y_test : ndarray, optional
        The testing labels if `merge_train_test=False`.
    extras : dict, optional
        The optional extras if `return_extras=True`.

    See Also
    --------
    load_dataset : load a named dataset
    """
    return load_dataset(
        "SyntheticControl",
        repository="wildboar/ucr-tiny",
        merge_train_test=merge_train_test,
        progress=False,
    )


# noqa: H0002
def load_two_lead_ecg(merge_train_test=True):
    """
    Load the TwoLeadECG dataset.

    Parameters
    ----------
    merge_train_test : bool, optional
        Merge the existing training and testing partitions.

    Returns
    -------
    x : ndarray, optional
        The samples if `merge_train_test=False`.
    y : ndarray, optional
        The labels, if `merge_train_test=False`.
    x_train : ndarray, optional
        The training samples if `merge_train_test=False`.
    x_test : ndarray, optional
        The testing samples if `merge_train_test=False`.
    y_train : ndarray, optional
        The training labels if `merge_train_test=False`.
    y_test : ndarray, optional
        The testing labels if `merge_train_test=False`.
    extras : dict, optional
        The optional extras if `return_extras=True`.

    See Also
    --------
    load_dataset : load a named dataset
    """
    return load_dataset(
        "TwoLeadECG",
        repository="wildboar/ucr-tiny",
        merge_train_test=merge_train_test,
        progress=False,
    )


# noqa: H0002
def load_gun_point(merge_train_test=True):
    """
    Load the GunPoint dataset.

    Parameters
    ----------
    merge_train_test : bool, optional
        Merge the existing training and testing partitions.

    Returns
    -------
    x : ndarray, optional
        The samples if `merge_train_test=False`.
    y : ndarray, optional
        The labels, if `merge_train_test=False`.
    x_train : ndarray, optional
        The training samples if `merge_train_test=False`.
    x_test : ndarray, optional
        The testing samples if `merge_train_test=False`.
    y_train : ndarray, optional
        The training labels if `merge_train_test=False`.
    y_test : ndarray, optional
        The testing labels if `merge_train_test=False`.
    extras : dict, optional
        The optional extras if `return_extras=True`.

    See Also
    --------
    load_dataset : load a named dataset
    """
    return load_dataset(
        "GunPoint",
        repository="wildboar/ucr-tiny",
        merge_train_test=merge_train_test,
        progress=False,
    )


def load_datasets(
    repository="wildboar/ucr",
    *,
    collection=None,
    cache_dir=None,
    create_cache_dir=True,
    progress=True,
    force=False,
    filter=None,
    **kwargs,
):
    """
    Load all datasets as a generator.

    Parameters
    ----------
    repository : str
        The repository string.
    collection : str, optional
        A collection of named datasets.
    cache_dir : str, optional
        The cache directory for downloaded dataset repositories.
    create_cache_dir : bool, optional
        Create the cache directory if it does not exist.
    progress : bool, optional
        If progress indicator is shown while downloading the repository.
    force : bool, optional
        Force re-download of cached repository.
    filter : str, dict, list or callable, optional
        Filter the datasets

        - if callable, only yield those datasets for which the callable returns
          `True`. ``f(dataset, x, y) -> bool``

        - if dict, filter based on the keys and values, where keys are
          attributes and values comparison specs

        - if list, filter based on conjunction of attribute comparisons

        - if str, filter based on attribute comparison

        Read more in the :ref:`User guide <datasets_filter>`.

        .. warning::

            If the parameter ``merge_train_test`` is ``False``, the filter is
            applied on the *training* part of the data.
    kwargs : dict
        Optional arguments to :func:`~wildboar.datasets.load_dataset`.

    Yields
    ------
    name : str
        The dataset name

    dataset : list
        Depends on the ``kwargs``.

        - If ``merge_train_test=True`` (default), dataset is a tuple of
          ``(x, y)``.
        - If ``merge_train_test=False``, dataset is a tuple of
          ``(x_train, x_test, y_train, y_test)``.
        - If ``return_extras=True``, the last element of the tuple contains
          optional extra (or ``None``).

    Examples
    --------
    Load all datasets in a repository:

    >>> for dataset, (x, y) in load_datasets(repository='wildboar/ucr-tiny'):
    ...     print(dataset, x.shape, y.shape)
    ...
    Beef (60, 470) (60,)
    Coffee (56, 286) (56,)
    GunPoint (200, 150) (200,)
    SyntheticControl (600, 60) (600,)
    TwoLeadECG (1162, 82) (1162,)

    Print the names of datasets with more than 200 samples

    >>> for dataset, (x, y) in load_datasets(
    ...    repository='wildboar/ucr-tiny', filter={"n_samples": ">200"}
    ... ):
    ...     print(dataset)
    SyntheticControl
    TwoLeadECG

    >>> for dataset, (x, y) in load_datasets(
    ...    repository='wildboar/ucr-tiny', filter="n_samples>200"
    ... ):
    ...     print(dataset)
    SyntheticControl
    TwoLeadECG
    """
    for dataset in list_datasets(
        repository=repository,
        collection=collection,
        cache_dir=cache_dir,
        create_cache_dir=create_cache_dir,
        progress=progress,
        force=force,
    ):
        if kwargs.get("merge_train_test", True):
            data = load_dataset(dataset, repository=repository, **kwargs)
            x = data[0]
            y = data[1]
        else:
            data = load_dataset(dataset, repository=repository, **kwargs)

            # We extract the training part to apply the filter.
            x = data[0]  # x_train
            y = data[2]  # y_train

        if filter is None:
            yield dataset, data
        elif hasattr(filter, "__call__") and filter(dataset, x, y):
            yield dataset, data
        elif make_filter(filter)(dataset, x, y):
            yield dataset, data


def load_dataset(  # noqa: PLR0912
    name,
    *,
    repository="wildboar/ucr",
    dtype=float,
    preprocess=None,
    contiguous=True,
    merge_train_test=True,
    cache_dir=None,
    create_cache_dir=True,
    progress=True,
    return_extras=False,
    force=False,
    refresh=False,
    timeout=None,
):
    """
    Load a dataset from a repository.

    Parameters
    ----------
    name : str
        The name of the dataset to load.
    repository : str, optional
        The data repository formatted as
        ``{repository}/{bundle}[:{version}][:{tag}]``. Read more in the
        :ref:`User guide <guide-datasets-repositories>`.
    dtype : dtype, optional
        The data type of x (train and test).
    preprocess : str, list or callable, optional
        Preprocess the dataset

        - if str, use named preprocess function (see ``preprocess._PREPROCESS.keys()``
          for valid keys).
        - if callable, function taking a np.ndarray and returns the
          preprocessed dataset.
        - if list, a list of callable or str.
    contiguous : bool, optional
        Ensure that the returned dataset is memory contiguous.
    merge_train_test : bool, optional
        Merge the existing training and testing partitions.
    cache_dir : str, optional
        The directory where downloaded files are cached.
    create_cache_dir : bool, optional
        Create cache directory if missing (default=True).
    progress : bool, optional
        Show a progress bar while downloading a bundle.
    return_extras : bool, optional
        Return optional extras.

        .. versionadded:: 1.1
    force : bool, optional
        Force re-download of already cached bundle.

        .. versionadded:: 1.0.4
    refresh : bool, optional
        Refresh the repository.

        .. versionadded:: 1.1
    timeout : float, optional
        Timeout for json request.

        .. versionadded:: 1.1

    Returns
    -------
    x : ndarray, optional
        The samples if `merge_train_test=False`.
    y : ndarray, optional
        The labels, if `merge_train_test=False`.
    x_train : ndarray, optional
        The training samples if `merge_train_test=False`.
    x_test : ndarray, optional
        The testing samples if `merge_train_test=False`.
    y_train : ndarray, optional
        The training labels if `merge_train_test=False`.
    y_test : ndarray, optional
        The testing labels if `merge_train_test=False`.
    extras : dict, optional
        The optional extras if `return_extras=True`.

    Examples
    --------
    Load a dataset from the default repository

    >>> x, y = load_dataset("SyntheticControl")
    >>> x.shape
    (600, 60)

    or if original training and testing splits are to be preserved

    >>> x_train, x_test, y_train, y_test = load_dataset(
    ...     "SyntheticControl", merge_train_test=False
    ... )

    or for a specific version of the dataset

    >>> x_train, x_test, y_train, y_test = load_dataset(
    ...     "SyntheticControl",
    ...     repository='wildboar/ucr-tiny:1.0.1',
    ...     merge_train_test=False,
    ... )
    """
    (
        repository_name,
        bundle_name,
        bundle_tag,
    ) = _split_repo_bundle(repository)
    cache_dir = cache_dir or _default_cache_dir()
    repository = get_repository(repository_name)

    if refresh:
        repository.refresh(timeout)

    ret_val = []
    x, y, n_train_samples, extras = repository.load_dataset(
        bundle_name,
        name,
        tag=bundle_tag,
        cache_dir=cache_dir,
        force=force,
        create_cache_dir=create_cache_dir,
        progress=progress,
    )
    if preprocess is None:

        def preprocess(identity):
            return identity

    if dtype:
        x = x.astype(dtype)

    if isinstance(preprocess, str):
        preprocess = check_option(_PREPROCESS, preprocess, "preprocess")
    if isinstance(preprocess, list):
        op = []
        for i, p in enumerate(preprocess):
            if isinstance(p, str):
                op.append(check_option(_PREPROCESS, p, "preprocess[%d]" % i))
            elif callable(p):
                op.append(p)
            else:
                raise TypeError(
                    "preprocess[%d] must be callable or str, not %r." % (i, p)
                )

        def preprocess(x):
            for o in op:
                x = o(x)
            return x

    elif not hasattr(preprocess, "__call__"):
        raise TypeError(
            "preprocess must be str, callable or list, not %r." % preprocess
        )

    x = preprocess(x)
    if merge_train_test:
        ret_val.append(x)
        ret_val.append(y)
    else:
        if n_train_samples == x.shape[0]:
            raise ValueError(
                "The dataset %s does not have a training testing split. "
                "Set merge_train_test=True to return the full dataset." % name
            )

        ret_val.append(x[:n_train_samples])
        ret_val.append(x[n_train_samples:])
        if y is not None:
            ret_val.append(y[:n_train_samples])
            ret_val.append(y[n_train_samples:])
        else:
            ret_val.append(None)
            ret_val.append(None)
    if return_extras:
        ret_val.append(extras)

    if contiguous:
        return [
            np.ascontiguousarray(x)
            if isinstance(x, np.ndarray)
            else {k: np.ascontiguousarray(v) for k, v in extras.items()}
            for x in ret_val
        ]
    else:
        return ret_val


# noqa: H0002
def list_datasets(
    repository="wildboar/ucr",
    *,
    collection=None,
    cache_dir=None,
    create_cache_dir=True,
    progress=True,
    force=False,
    refresh=False,
    timeout=None,
):
    """
    List the datasets in the repository.

    Parameters
    ----------
    repository : str or Bundle, optional
        The data repository

        - if str load a named bundle, format `{repository}/{bundle}`.
    collection : str, optional
        A collection of named datasets.
    cache_dir : str, optional
        The directory where downloaded files are cached (default='wildboar_cache').
    create_cache_dir : bool, optional
        Create cache directory if missing (default=True).
    progress : bool, optional
        Show a progress bar while downloading a bundle.
    force : bool, optional
        Force re-download of cached bundle.
    refresh : bool, optional
        Refresh the repository.

        .. versionadded:: 1.1
    timeout : float, optional
        Timeout for json request.

        .. versionadded:: 1.1

    Returns
    -------
    set
        A set of dataset names.
    """
    (
        repository_name,
        bundle_name,
        bundle_tag,
    ) = _split_repo_bundle(repository)
    cache_dir = cache_dir or _default_cache_dir()
    repository = get_repository(repository_name)
    if refresh:
        repository.refresh(timeout)
    return repository.list_datasets(
        bundle_name,
        collection=collection,
        tag=bundle_tag,
        cache_dir=cache_dir,
        create_cache_dir=create_cache_dir,
        progress=progress,
        force=force,
    )


# noqa: H0002
def clear_cache(repository=None, *, cache_dir=None, keep_last_version=True):
    """
    Clear the cache by deleting cached datasets.

    Parameters
    ----------
    repository : str, optional
        The name of the repository to clear cache.

        - if None, clear cache of all repositories.
    cache_dir : str, optional
        The cache directory.
    keep_last_version : bool, optional
        If true, keep the latest version of each repository.
    """
    cache_dir = cache_dir or _default_cache_dir()
    if repository is None:
        for repo in _REPOSITORIES:
            repo.clear_cache(cache_dir=cache_dir, keep_last_version=keep_last_version)
    else:
        get_repository(repository).clear_cache(
            cache_dir=cache_dir, keep_last_version=keep_last_version
        )


# noqa: H0002
def get_repository(repository):
    """
    Get repository by name.

    Parameters
    ----------
    repository : str
        Repository name.

    Returns
    -------
    Repository
        A repository.
    """
    return _REPOSITORIES[repository]


# noqa: H0002
def install_repository(repository, *, refresh=True, timeout=None, cache_dir=None):
    """
    Install repository.

    Parameters
    ----------
    repository : str or Repository
        A repository.
    refresh : bool, optional
        Refresh the repository.

        .. versionadded:: 1.1
    timeout : float, optional
        Timeout for json request.

        .. versionadded:: 1.1
    cache_dir : str, optional
        Cache directory.

        .. versionadded:: 1.1
    """
    if isinstance(repository, str):
        repository = JSONRepository(repository)

    cache_dir = cache_dir or _default_cache_dir()
    _REPOSITORIES.install(
        repository, refresh=refresh, timeout=timeout, cache_dir=cache_dir
    )


# noqa: H0002
def refresh_repositories(repository=None, *, timeout=None, cache_dir=None):
    """
    Refresh the installed repositories.

    Parameters
    ----------
    repository : str, optional
        The repository. None means all repositories.
    timeout : float, optional
        Timeout for request.

        .. versionadded:: 1.1
    cache_dir : str, optional
        Cache directory.

        .. versionadded:: 1.1
    """
    cache_dir = cache_dir or _default_cache_dir()
    _REPOSITORIES.refresh(repository=repository, timeout=timeout, cache_dir=cache_dir)


# noqa: H0002
def get_bundles(repository, *, refresh=False, timeout=None):
    """
    Get all bundles in the repository.

    Parameters
    ----------
    repository : str
        Name of the repository.
    refresh : bool, optional
        Refresh the repository.

        .. versionadded:: 1.1
    timeout : float, optional
        Timeout for json request.

        .. versionadded:: 1.1

    Returns
    -------
    dict
        A dict of key Bundle pairs.
    """
    repository = _REPOSITORIES[repository]
    if refresh:
        repository.refresh(timeout)

    return repository.get_bundles()


def list_bundles(repository, *, refresh=False, timeout=None):
    """
    Get a list of all bundle names in the specified repository.

    Parameters
    ----------
    repository : str
        The name of the repository.
    refresh : bool, optional
        Refresh the repository.

        .. versionadded:: 1.1
    timeout : float, optional
        Timeout for json request.

        .. versionadded:: 1.1

    Returns
    -------
    list
        A list of bundle names.

    Examples
    --------
    >>> from wildboar.datasets import list_bundles
    >>> list_bundles("wildboar")
    ["ucr", "ucr-tiny", ...]
    """
    return sorted(get_bundles(repository, refresh=refresh, timeout=timeout).keys())


def list_collections(repository):
    """
    List the collections of the repository.

    Parameters
    ----------
    repository : str or Bundle, optional
        The data repository

        - if str load a named bundle, format {repository}/{bundle}.

    Returns
    -------
    collections
        A list of collections.

    Examples
    --------
    >>> from wildboar.datasets import list_collections
    >>> list_collections("wildboar/ucr")
    ["bake-off", ...]
    """
    (
        repository_name,
        bundle_name,
        _,
    ) = _split_repo_bundle(repository)
    repository = get_repository(repository_name)
    collections = repository.get_bundle(bundle_name).collections
    return sorted(collections.keys()) if collections is not None else []


def list_repositories(*, refresh=False, timeout=None, cache_dir=None):
    """
    List the key of all installed repositories.

    Parameters
    ----------
    refresh : bool, optional
        Refresh all repositories.

        .. versionadded:: 1.1
    timeout : float, optional
        Timeout for json request.

        .. versionadded:: 1.1
    cache_dir : str, optional
        Cache directory.

        .. versionadded:: 1.1

    Returns
    -------
    repositories
        A list of installed repositories.

    Examples
    --------
    >>> from wildboar.datasets import list_repositories
    >>> list_repositories()
    ["wildboar", ...]

    We can also refresh the repositories, to load any newly added
    but still pending repositories.

    >>> list_repositories(refresh=True)
    """
    if refresh:
        cache_dir = cache_dir or _default_cache_dir()
        refresh_repositories(timeout=timeout, cache_dir=cache_dir)
    return sorted([repo.name for repo in _REPOSITORIES])


_CACHE_DIR = None
_REPOSITORIES = RepositoryCollection()


# Install the default 'wildboar' repository
def _get_dataset_version():
    from sklearn.utils.fixes import parse_version

    from .. import __version__

    v = parse_version(__version__)
    if v.dev is not None:
        return "master"
    else:
        return "%d.%d" % (v.major, v.minor)


install_repository(
    "https://datasets.wildboar.dev/%s/repo.json" % _get_dataset_version(),
    refresh=False,
)



"""Explanation methods for classifiers and regressors."""

# Authors: Isak Samsten
# License: BSD 3 clause

from ._importance import (
    AmplitudeImportance,
    FrequencyImportance,
    IntervalImportance,
    ShapeletImportance,
    plot_importances,
)

__all__ = [
    "IntervalImportance",
    "AmplitudeImportance",
    "ShapeletImportance",
    "FrequencyImportance",
    "plot_importances",
]



# Authors: Isak Samsten
# License: BSD 3 clause

import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_equal

from wildboar.datasets import load_dataset
from wildboar.datasets.outlier import minority_outliers


@pytest.mark.parametrize(
    "n_outliers, expected_n_outliers, "
    "expected_first_outlier_sample, expected_outlier_norm",
    [
        (0.05, 5, 18, 27.294688068996287),
        (0.1, 10, 18, 38.60051804670847),
        (0.2, 20, 18, 54.58937618813437),
    ],
)
def test_minority_outliers(
    n_outliers,
    expected_n_outliers,
    expected_first_outlier_sample,
    expected_outlier_norm,
):
    x, y = load_dataset("GunPoint", repository="wildboar/ucr-tiny")
    x_outlier, y_outlier = minority_outliers(
        x, y, n_outliers=n_outliers, random_state=123
    )
    assert_outliers(
        x,
        x_outlier,
        y_outlier,
        expected_n_outliers=expected_n_outliers,
        expected_first_outlier_sample=expected_first_outlier_sample,
        expected_outlier_norm=expected_outlier_norm,
    )


def assert_outliers(
    x,
    x_outlier,
    y_outlier,
    *,
    expected_n_outliers,
    expected_first_outlier_sample,
    expected_outlier_norm
):
    actual_n_outliers = np.where(y_outlier == -1)[0].shape[0]
    actual_fist_outlier_sample = (np.sum(x_outlier[0] - x, axis=1) == 0).argmax()
    actual_outlier_norm = np.linalg.norm(x_outlier[np.where(y_outlier == -1)])
    assert_almost_equal(actual_outlier_norm, expected_outlier_norm)
    assert_equal(actual_fist_outlier_sample, expected_first_outlier_sample)
    assert_equal(actual_n_outliers, expected_n_outliers)



# cython: language_level=3

# Authors: Isak Samsten
# License: BSD 3 clause

cdef struct Deque:
    Py_ssize_t *queue
    Py_ssize_t size
    Py_ssize_t capacity
    Py_ssize_t front
    Py_ssize_t back

cdef void deque_init(Deque *c, Py_ssize_t capacity) noexcept nogil

cdef void deque_reset(Deque *c) noexcept nogil

cdef void deque_destroy(Deque *c) noexcept nogil

cdef void deque_push_back(Deque *c, Py_ssize_t v) noexcept nogil

cdef void deque_pop_front(Deque *c) noexcept nogil

cdef void deque_pop_back(Deque *c) noexcept nogil

cdef Py_ssize_t deque_front(Deque *c) noexcept nogil

cdef Py_ssize_t deque_back(Deque *c) noexcept nogil

cdef bint deque_empty(Deque *c) noexcept nogil

cdef Py_ssize_t deque_size(Deque *c) noexcept nogil

cdef void find_min_max(
    const double *T,
    Py_ssize_t length,
    Py_ssize_t r,
    double *lower,
    double *upper,
    Deque *dl,
    Deque *du,
) noexcept nogil

cdef double scaled_dtw_subsequence_distance(
    const double *S,
    Py_ssize_t s_length,
    double s_mean,
    double s_std,
    const double *T,
    Py_ssize_t t_length,
    Py_ssize_t r,
    double *X_buffer,
    double *cost,
    double *cost_prev,
    double *s_lower,
    double *s_upper,
    double *t_lower,
    double *t_upper,
    double *cb,
    double *cb_1,
    double *cb_2,
    Py_ssize_t *index,
) noexcept nogil

cdef double dtw_subsequence_distance(
    const double *S,
    Py_ssize_t s_length,
    const double *T,
    Py_ssize_t t_length,
    Py_ssize_t r,
    double *cost,
    double *cost_prev,
    double *weight_vector,
    Py_ssize_t *index,
) noexcept nogil

cdef double dtw_distance(
    const double *X,
    Py_ssize_t x_length,
    const double *Y,
    Py_ssize_t y_length,
    Py_ssize_t r,
    double *cost,
    double *cost_prev,
    double *weight_vector,
    double min_dist,
) noexcept nogil

cdef double adtw_distance(
    const double *X,
    Py_ssize_t x_length,
    const double *Y,
    Py_ssize_t y_length,
    Py_ssize_t r,
    double *cost,
    double *cost_prev,
    double penalty,
    double min_dist,
) noexcept nogil

cdef double lcss_distance(
    const double *X,
    Py_ssize_t x_length,
    const double *Y,
    Py_ssize_t y_length,
    Py_ssize_t r,
    double threshold,
    double *cost,
    double *cost_prev,
    double *weight_vector,
    double min_dist,
) noexcept nogil

cdef double erp_distance(
    const double *X,
    Py_ssize_t x_length,
    const double *Y,
    Py_ssize_t y_length,
    Py_ssize_t r,
    double g,
    double *gX,
    double *gY,
    double *cost,
    double *cost_prev,
    double min_dist,
) noexcept nogil

cdef double edr_distance(
    const double *X,
    Py_ssize_t x_length,
    const double *Y,
    Py_ssize_t y_length,
    Py_ssize_t r,
    double threshold,
    double *cost,
    double *cost_prev,
    double *weight_vector,
    double min_dist,
) noexcept nogil

cdef double msm_distance(
    const double *X,
    Py_ssize_t x_length,
    const double *Y,
    Py_ssize_t y_length,
    Py_ssize_t r,
    double c,
    double *cost,
    double *cost_prev,
    double *cost_y,
    double min_dist,
) noexcept nogil

cdef double twe_distance(
    const double *X,
    Py_ssize_t x_length,
    const double *Y,
    Py_ssize_t y_length,
    Py_ssize_t r,
    double penalty,
    double stiffness,
    double *cost,
    double *cost_prev,
    double min_dist,
) nogil
